# -*- coding: utf-8 -*-
"""
illumination fields for torchgdm
"""
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.env.base_classes import IlluminationfieldBase
from torchgdm.tools.geometry import rotation_x, rotation_y, rotation_z


# ==============================================================================
# field generator classes
# ==============================================================================
class NullField(IlluminationfieldBase):
    """Zero-Field

    all additional kwargs are ignored

    """

    __name__ = "null field"

    def __init__(
        self,
        device: torch.device = None,
    ):
        super().__init__(device=device)
        self.n_dim = 3

    def set_device(self, device):
        super().set_device(device)

    def get_efield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate illumination field at position(s) `r_probe`"""
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)
        assert r_probe.shape[1] == 3

        return torch.zeros((len(r_probe), 3), dtype=DTYPE_COMPLEX, device=self.device)

    def get_hfield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate illumination field at position(s) `r_probe`"""
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)
        assert r_probe.shape[1] == 3

        return torch.zeros((len(r_probe), 3), dtype=DTYPE_COMPLEX, device=self.device)


class _PlaneWaveNormalInc(IlluminationfieldBase):
    """Simple plane wave in homogeneous environment, normal incidence (along z)

    The polarization state can be defined in different ways. The s- and p-polarization
    complex amplitudes can be defined with a relative phase using. Alternatively a
    relative phase (in rad) can be given for the s-polarization component.

    Parameters
    ----------
    e0x : complex, default: 1.0
        E_x constant amplitude factor (complex for possible phase wrt E_y)

    e0y : complex, default: 0.0
        E_y constant amplitude factor (complex for possible phase wrt E_x)

    k_z_sign : int, default: -1
        sign of z-wavenumber (incidence direction).
        +1: towards increasing z, -1: towards smaller z (default)

    """

    __name__ = "normal incidence plane wave"

    def __init__(
        self,
        e0x: complex = 1.0,
        e0y: complex = 0.0,
        k_z_sign: float = -1.0,
        absolute_phase: float = 0.0,
        device: torch.device = None,
    ):
        super().__init__(device=device)
        self.n_dim = 3

        self.e0x = e0x
        self.e0y = e0y
        self.k_z_sign = k_z_sign
        self.absolute_phase = absolute_phase
        if self.k_z_sign not in [-1, 1]:
            raise ValueError("planewave: kSign must be either +1 or -1!")

        warnings.warn(
            "`_IlluminationPlaneWave_simple` is deprecated. Will be removed soon."
        )

    def set_device(self, device):
        super().set_device(device)

    def __repr__(self, verbose=False):
        out_str = " ------ {} -------".format(self.__name__)
        out_str += "\n - pol: {} ex + {} ey ".format(self.e0x, self.e0y)
        out_str += "\n - sign(k_z): {}".format(self.k_z_sign)

        return out_str

    def get_info(self):
        return dict(
            name="plane wave",
            e0x=self.e0x,
            e0y=self.e0y,
            k_z_sign=self.k_z_sign,
            abs_phase=self.absolute_phase,
        )

    def _eval_field(
        self,
        r_probe: torch.Tensor,
        wavelength: float,
        environment,
    ):
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=r_probe
        )
        n_env = eps_env**0.5

        k_z = self.k_z_sign * n_env * (2 * torch.pi / wavelength)  # incidence along z

        field_x = torch.ones(len(r_probe), dtype=DTYPE_COMPLEX, device=self.device)
        field_y = torch.ones(len(r_probe), dtype=DTYPE_COMPLEX, device=self.device)
        field_z = torch.zeros(len(r_probe), dtype=DTYPE_COMPLEX, device=self.device)

        phase_factor = torch.exp(1j * (k_z * r_probe[:, 2] + self.absolute_phase))
        field_x *= self.e0x * phase_factor
        field_y *= self.e0y * phase_factor

        return field_x, field_y, field_z

    def get_efield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate illumination electric field at position(s) `r_probe`"""
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        assert r_probe.shape[1] == 3

        ex, ey, ez = self._eval_field(
            r_probe=r_probe,
            wavelength=wavelength,
            environment=environment,
        )

        return torch.stack((ex, ey, ez), dim=1)

    def get_hfield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate illumination magnetic field at position(s) `r_probe`"""
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        hy, hx, hz = self._eval_field(
            r_probe=r_probe,
            wavelength=wavelength,
            environment=environment,
        )
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=r_probe
        )
        n_env = eps_env**0.5
        return torch.stack((hx, -1 * hy, hz), dim=1) * n_env.unsqueeze(1)


class PlaneWave(IlluminationfieldBase):
    """3D plane wave in homogeneous environment"""

    __name__ = "free space plane wave 3D"

    def __init__(
        self,
        e0s=0.0,
        e0p=1.0,
        inc_angle=0.0,
        phase_e0s=0.0,
        phase_global=0,
        inc_plane="xz",
        device: torch.device = None,
    ):
        """plane wave in homogeneous 3D environment

        The polarization state can be defined in different ways. The s- and p-polarization
        complex amplitudes can be defined with a relative phase using. Alternatively a
        relative phase (in rad) can be given for the s-polarization component.

        Args:
            e0s (complex, optional): s-polarization amplitude factor. Defaults to 0.0.
            e0p (complex, optional): s-polarization amplitude factor. Defaults to 1.0.
            inc_angle (float, optional): incident angle in rad. 0:k=-ez, pi/2:k=ex, pi:k=+ez. Defaults to 0.0.
            inc_plane (str, optional): Either "xz" or "yz". Defaults to "xz".
            phase_e0s (float, optional): additional phase of s-amplitude vector (in rad). Defaults to 0.0.
            phase_global (int, optional): additional global phase (in rad). Defaults to 0.
            device (torch.device, optional): Defaults to "cpu".
        """
        super().__init__(device=device)
        self.n_dim = 3

        self.e0s = torch.as_tensor(e0s, dtype=DTYPE_COMPLEX, device=device)
        self.e0p = torch.as_tensor(e0p, dtype=DTYPE_COMPLEX, device=device)
        # in rad (xz inc_plane: 0=-ez, pi/2=ex, pi=ez)
        self.inc_angle = torch.as_tensor(inc_angle, dtype=DTYPE_FLOAT, device=device)
        self.phase_e0s = torch.as_tensor(phase_e0s, dtype=DTYPE_FLOAT, device=device)
        self.phase_global = torch.as_tensor(
            phase_global, dtype=DTYPE_FLOAT, device=device
        )

        self.inc_plane = inc_plane.lower()  # 'xz' or 'yz'
        if self.inc_plane == "xz":
            self.k_vec_unit = torch.as_tensor(
                [-torch.sin(self.inc_angle), 0.0, torch.cos(self.inc_angle)],
                dtype=DTYPE_COMPLEX,
                device=self.device,
            )
            self.e0s_vec_unit = torch.as_tensor(
                [0.0, 1.0, 0.0],
                dtype=DTYPE_COMPLEX,
                device=self.device,
            )
            self.e0p_vec_unit = torch.as_tensor(
                [-torch.cos(self.inc_angle), 0.0, -torch.sin(self.inc_angle)],
                dtype=DTYPE_COMPLEX,
                device=self.device,
            )
        elif self.inc_plane == "yz":
            self.k_vec_unit = torch.as_tensor(
                [0.0, -torch.sin(self.inc_angle), torch.cos(self.inc_angle)],
                dtype=DTYPE_COMPLEX,
                device=self.device,
            )
            self.e0s_vec_unit = torch.as_tensor(
                [-1.0, 0.0, 0.0],
                dtype=DTYPE_COMPLEX,
                device=self.device,
            )
            self.e0p_vec_unit = torch.as_tensor(
                [0.0, -torch.cos(self.inc_angle), -torch.sin(self.inc_angle)],
                dtype=DTYPE_COMPLEX,
                device=self.device,
            )
        else:
            raise ValueError(
                "planewave currently supports only 'xz' and 'yz' incident planes."
            )

    def set_device(self, device):
        super().set_device(device)
        self.e0s = torch.as_tensor(self.e0s, device=device)
        self.e0p = torch.as_tensor(self.e0p, device=device)
        self.inc_angle = torch.as_tensor(self.inc_angle, device=device)
        self.phase_e0s = torch.as_tensor(self.phase_e0s, device=device)
        self.phase_global = torch.as_tensor(self.phase_global, device=device)
        self.k_vec_unit = torch.as_tensor(self.k_vec_unit, device=device)
        self.e0s_vec_unit = torch.as_tensor(self.e0s_vec_unit, device=device)
        self.e0p_vec_unit = torch.as_tensor(self.e0p_vec_unit, device=device)

    def __repr__(self, verbose=False):
        out_str = " ------ {} -------".format(self.__name__)
        out_str += "\n - pol: {} e_p + {} e_s ".format(self.e0p, self.e0s)
        out_str += "\n - additional e_s phase (rad): {}".format(self.phase_e0s)
        out_str += "\n - incident angle (rad): {}".format(self.inc_angle)
        out_str += "\n - plane of incidence  : {}".format(self.inc_plane)

        return out_str

    def get_info(self):
        """Get info about field config. as dictionary"""
        return dict(
            name="plane wave",
            e0s=self.e0s,
            e0p=self.e0p,
            inc_angle=self.inc_angle,
            phase_e0s=self.phase_e0s,
            phase_global=self.phase_global,
        )

    def _eval_field(
        self,
        r_probe: torch.Tensor,
        wavelength: float,
        environment,
    ):
        # calculate wave vector and phase factor
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=r_probe
        )
        n_env = eps_env**0.5

        k0 = 2 * torch.pi / torch.as_tensor(wavelength, device=self.device)
        k_vec = self.k_vec_unit.unsqueeze(0) * k0 * n_env.unsqueeze(1)
        kr = (k_vec * r_probe).sum(axis=1)  # k dot r for each vector pair
        phase_vec = torch.exp(1j * (kr + self.phase_global))

        # 2D simulations: test if y-component of k-vector matches environment
        if hasattr(environment, "get_k0_y"):
            k0_y_field = torch.as_tensor(
                torch.abs(self.k_vec_unit[1]) * k0, dtype=DTYPE_COMPLEX
            )
            k0_y_env = torch.as_tensor(
                environment.get_k0_y(wavelength), dtype=DTYPE_COMPLEX
            )
            if not torch.allclose(k0_y_field, k0_y_env, atol=1e-6, rtol=1e-6):
                raise ValueError(
                    "k-vector component along y must match `k0_y` of the 2D environment class. "
                    + "But k0_y field: {} != k0_y environment: {}".format(
                        k0_y_field, k0_y_env
                    )
                    + "Check and correct the incident angle configuration."
                )

        # calculate s and p field amplitudes
        es = (
            self.e0s_vec_unit.unsqueeze(0)
            * phase_vec.unsqueeze(1)
            * torch.exp(
                1j * self.phase_e0s
            )  # optional phase of s-polarization component
        )
        ep = self.e0p_vec_unit.unsqueeze(0) * phase_vec.unsqueeze(1)

        return es, ep

    def get_efield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's electric field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: E-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        es, ep = self._eval_field(
            r_probe=r_probe,
            wavelength=wavelength,
            environment=environment,
        )

        return self.e0s * es + self.e0p * ep

    def get_hfield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's magnetic field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: H-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        es, ep = self._eval_field(
            r_probe=r_probe,
            wavelength=wavelength,
            environment=environment,
        )
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=r_probe
        )
        n_env = eps_env**0.5
        return (-self.e0p * es + self.e0s * ep) * n_env.unsqueeze(1)


class GaussianParaxial(IlluminationfieldBase):
    """Paraxial Gaussian beam with tight focus correction

    homogeneous environment, 'xz' incident plane.

    The polarization state is defined by Ex and Ey complex amplitude
    as well as a relative phase between x and y (optional).

    Parameters
    ----------
    e0x : complex, default: 1.0
        E_x constant amplitude factor (complex for possible phase wrt E_y)

    e0y : complex, default: 0.0
        E_y constant amplitude factor (complex for possible phase wrt E_x)

    k_z_sign : int, default: -1
        sign of z-wavenumber (incidence direction).
        +1: towards increasing z, -1: towards smaller z (default)

    Notes
    -----
     - paraxial correction, see:
       Novotny & Hecht. "Principles of nano-optics". Cambridge University Press (2006)

    """

    __name__ = "paraxial focused Gaussian beam"

    def __init__(
        self,
        e0s=0.0,
        e0p=1.0,
        r_focus=torch.as_tensor([0.0, 0.0, 0.0]),
        NA=None,
        waist=None,
        inc_angle=0.0,
        phase_e0s=0.0,
        phase_global=0,
        correction_div_e=True,
        device: torch.device = None,
    ):
        """focused paraxial Gaussian in homogeneous 3D environment, in "xz" incident plane

        tight focus paraxial correction, see:
        Novotny & Hecht. "Principles of nano-optics". Cambridge University Press (2006)

        The polarization state can be defined in different ways. The s- and p-polarization
        complex amplitudes can be defined directly, or a relative phase for the s-component
        can be given.

        Args:
            e0s (complex, optional): s-polarization amplitude factor. Defaults to 0.0.
            e0p (complex, optional): s-polarization amplitude factor. Defaults to 1.0.
            r_focus (float torch.Tensor, optional): focal position (in nm). Defaults to torch.as_tensor([0.0, 0.0, 0.0]).
            NA (float, optional): numerical aperture used for focal spotsize calculation. If None, use explicit `waist`. Defaults to None.
            waist (float, optional): explicit, constant Gausian waist (in nm). Defaults to None.
            inc_angle (float, optional): incident angle in rad. 0:k=-ez, pi:k=+ez. Defaults to 0.0.
            inc_plane (str, optional): either "xz" or "yz". Defaults to "xz".
            phase_e0s (float, optional): additional phase of s-amplitude vector (in rad). Defaults to 0.0.
            phase_global (int, optional): additional global phase (in rad). Defaults to 0.
            correction_div_e (bool, optional): whether to use paraxial tight focus correction term. Defaults to True.
            device (torch.device, optional): Defaults to "cpu".
        """
        super().__init__(device=device)
        self.n_dim = 3

        self.e0s = torch.as_tensor(e0s, dtype=DTYPE_COMPLEX, device=device)
        # sign -1 for p-polarization for orthonormal system handedness convention
        self.e0p = torch.as_tensor(e0p, dtype=DTYPE_COMPLEX, device=device)
        self.r_focus = torch.as_tensor(r_focus, dtype=DTYPE_FLOAT, device=device)
        self.inc_angle = torch.as_tensor(inc_angle, dtype=DTYPE_FLOAT, device=device)
        self.phase_e0s = torch.as_tensor(phase_e0s, dtype=DTYPE_FLOAT, device=device)
        self.phase_global = torch.as_tensor(
            phase_global, dtype=DTYPE_FLOAT, device=device
        )

        self.epsilon = 1e-5  # avoid div by zero at pi/2
        self.rot_y = rotation_y(self.inc_angle + self.epsilon, device=self.device)

        self.waist = waist
        self.NA = NA
        if self.waist is None and self.NA is None:
            raise ValueError(
                "Gaussian Beam Error: Either waist or NA must be given."
            )
        if self.waist is not None and self.NA is not None:
            warnings.warn("Both, waist and NA given. Ignoring `NA`.")
            self.NA = None
        if self.NA is not None:
            self.NA = torch.as_tensor(self.NA, device=device)
        if self.waist is not None:
            self.waist = torch.as_tensor(self.waist, device=device)

        self.correction_div_e = correction_div_e

    def set_device(self, device):
        super().set_device(device)
        self.e0s = torch.as_tensor(self.e0s, device=device)
        self.e0p = torch.as_tensor(self.e0p, device=device)
        self.r_focus = torch.as_tensor(self.r_focus, device=device)
        self.inc_angle = torch.as_tensor(self.inc_angle, device=device)
        self.phase_e0s = torch.as_tensor(self.phase_e0s, device=device)
        self.phase_global = torch.as_tensor(self.phase_global, device=device)
        self.rot_y = torch.as_tensor(self.rot_y, device=device)
        if self.NA is not None:
            self.NA = torch.as_tensor(self.NA, device=device)
        if self.waist is not None:
            self.waist = torch.as_tensor(self.waist, device=device)

    def __repr__(self, verbose=False):
        out_str = " ------ {} -------".format(self.__name__)
        out_str += "\n - pol: {} e_p + {} e_s ".format(self.e0p, self.e0s)
        out_str += "\n - additional e_s phase (rad): {}".format(self.phase_e0s)
        out_str += "\n - focal position (nm): {}".format(
            torch.round(self.r_focus, decimals=1)
        )
        if self.waist is None:
            out_str += "\n - numerical aperture: {}".format(
                torch.round(self.NA, decimals=4)
            )
        else:
            out_str += "\n - waist (Gaussian width): {}nm".format(
                torch.round(self.waist, decimals=1)
            )
        out_str += "\n - incident angle (rad): {}".format(self.inc_angle)
        out_str += "\n - plane of inclination  : xz"

        return out_str

    def get_info(self):
        """Get info about field config. as dictionary"""
        return dict(
            name="Gaussian",
            e0s=self.e0s,
            e0p=self.e0p,
            inc_angle=self.inc_angle,
            r_focus=self.r_focus,
            phase_e0s=self.phase_e0s,
            phase_global=self.phase_global,
        )

    def get_x(self):
        """return x position of focus"""
        return self.r_focus.squeeze()[0]

    def get_y(self):
        """return y position of focus"""
        return self.r_focus.squeeze()[1]

    def get_z(self):
        """return z position of focus"""
        return self.r_focus.squeeze()[2]

    # waist
    def _w(self, z, a_foc, w0):
        # eps: avoid div/0 at focus
        return w0 * torch.sqrt(1 + (z / (a_foc + self.epsilon)) ** 2)

    # curvature
    def _R(self, z, a_foc):
        # eps: avoid div/0 at focus
        return z * (1 + (a_foc / (z + self.epsilon)) ** 2) + self.epsilon

    # gouy-phase
    def _gouy(self, z, a_foc):
        return torch.arctan2(z, a_foc)

    def _eval_field(
        self,
        r_probe: torch.Tensor,
        wavelength: float,
        environment,
    ):
        # --- coordinate transform.
        # shift focus, rotate around Y-axis
        r_t = torch.clone(r_probe)
        r_t -= self.r_focus.unsqueeze(0)
        r_t = torch.matmul(r_t.unsqueeze(1), self.rot_y.unsqueeze(0))
        r_t = r_t[:, 0]
        if len(r_t.shape) == 1:
            r_t = r_t.unsqueeze(0)

        # relative coordinates
        x = r_t[:, 0]
        y = r_t[:, 1]
        z = r_t[:, 2]
        z += self.epsilon * (z == 0.0).type(torch.float32)
        r2_foc = x**2 + y**2

        # wavenumber
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=r_t
        )
        n_env = eps_env**0.5
        k_z = (
            n_env
            * 2
            * torch.pi
            / torch.as_tensor(wavelength, dtype=DTYPE_FLOAT, device=self.device)
        )

        # gaussian beam waist
        if self.waist is None:
            w0 = torch.as_tensor(
                2 * wavelength / (self.NA * torch.pi),
                dtype=DTYPE_FLOAT,
                device=self.device,
            )
        else:
            w0 = torch.as_tensor(self.waist, dtype=DTYPE_FLOAT, device=self.device)
        a_foc = torch.pi * w0**2 / wavelength
        waist = self._w(z, a_foc, w0)

        # normalized complex amplitude
        e_gaussian = ((w0 / waist) * torch.exp(-r2_foc / waist**2)) * torch.exp(
            1j
            * (
                k_z * z
                + k_z * (r2_foc / (2 * self._R(z, a_foc)))
                - self._gouy(z, a_foc)
            )
        )

        # optional phase on ey (after rotation this will be the s-polarization)
        ex = e_gaussian * torch.exp(1j * self.phase_global)
        ey = e_gaussian * torch.exp(1j * (self.phase_global + self.phase_e0s))

        # optionally: correct for divE=0
        if self.correction_div_e:
            ez = (-1j * 2 / (k_z * self._w(z, a_foc, w0) ** 2)) * (
                x * (ex * self.e0p) + y * (ey * self.e0s)
            )
        else:
            ez = torch.zeros_like(e_gaussian)

        # --- coordinate transform, xz scattering plane: fields rotate around y axis
        ef = torch.stack([ex, ey, ez], axis=1)
        ef = torch.matmul(
            ef.unsqueeze(1), self.rot_y.t().unsqueeze(0).to(dtype=DTYPE_COMPLEX)
        )
        ef = ef[:, 0]
        if len(ef.shape) == 1:
            ef = ef.unsqueeze(0)

        return ef[:, 0], ef[:, 1], ef[:, 2]

    def get_efield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's electric field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: E-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        ep, es, ek = self._eval_field(
            r_probe=r_probe,
            wavelength=wavelength,
            environment=environment,
        )

        return torch.stack([-self.e0p * ep, self.e0s * es, self.e0p * ek], axis=1)

    def get_hfield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's magnetic field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: H-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        ep, es, ek = self._eval_field(
            r_probe=r_probe,
            wavelength=wavelength,
            environment=environment,
        )
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=r_probe
        )
        n_env = eps_env**0.5

        return torch.stack(
            [-self.e0s * ep, -self.e0p * es, self.e0s * ek], axis=1
        ) * n_env.unsqueeze(1)


class ElectricDipole(IlluminationfieldBase):
    """electric point-dipole light source in homogeneous 3D environment"""

    __name__ = "electric dipole source"

    def __init__(
        self,
        r_source: torch.Tensor,
        p_source: torch.Tensor,
        device: torch.device = None,
    ):
        """electric point dipole in homogeneous environment

        Args:
            r_source (complex torch.Tensor): 3D position of dipole as [x, y, z]
            p_source (complex torch.Tensor): amplitude vector (3D) of the source's electric dipole moment as [p_x, p_y, p_z]
            device (torch.device, optional): Defaults to "cpu".
        """
        super().__init__(device=device)
        self.n_dim = 3

        self.r_source = torch.as_tensor(r_source, dtype=DTYPE_FLOAT, device=device)
        self.p_source = torch.as_tensor(p_source, dtype=DTYPE_COMPLEX, device=device)

        assert len(self.r_source) == 3
        assert len(self.r_source.shape) == 1
        assert len(self.p_source) == 3
        assert len(self.p_source.shape) == 1

    def set_device(self, device):
        super().set_device(device)
        self.r_source = self.r_source.to(device=device)
        self.p_source = self.p_source.to(device=device)

    def __repr__(self, verbose=False):
        out_str = " ------ {} -------".format(self.__name__)
        out_str += "\n - location (nm): {}".format(
            torch.round(self.r_source, decimals=1)
        )
        out_str += "\n - amplitude: {} ".format(self.p_source)

        return out_str

    def get_info(self):
        """Get info about field config. as dictionary"""
        return dict(
            name=self.__name__,
            r_source=self.r_source,
            p_source=self.p_source,
        )

    def get_x(self):
        """return x position of dipole"""
        return self.r_source.squeeze()[0]

    def get_y(self):
        """return y position of dipole"""
        return self.r_source.squeeze()[1]

    def get_z(self):
        """return z position of dipole"""
        return self.r_source.squeeze()[2]

    def get_efield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's electric field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: E-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        G_Ep = environment.get_G_Ep(r_probe, self.r_source.unsqueeze(0), wavelength)
        Ep = torch.matmul(G_Ep, self.p_source.unsqueeze(-1))[..., 0]
        return Ep

    def get_hfield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's magnetic field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: H-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        G_Hp = environment.get_G_Hp(r_probe, self.r_source.unsqueeze(0), wavelength)
        Hp = torch.matmul(G_Hp, self.p_source.unsqueeze(-1))[..., 0]
        return Hp


class MagneticDipole(IlluminationfieldBase):
    """magnetic point-dipole light source in homogeneous 3D environment"""

    __name__ = "magnetic dipole source"

    def __init__(
        self,
        r_source: torch.Tensor,
        m_source: torch.Tensor,
        device: torch.device = None,
    ):
        """magnetic point dipole in homogeneous environment

        Args:
            r_source (complex torch.Tensor): 3D position of dipole as [x, y, z]
            m_source (complex torch.Tensor): amplitude vector (3D) of the source's magnetic dipole moment as [p_x, p_y, p_z]
            device (torch.device, optional): Defaults to "cpu".
        """
        super().__init__(device=device)
        self.n_dim = 3

        self.r_source = torch.as_tensor(r_source, dtype=DTYPE_FLOAT, device=device)
        self.m_source = torch.as_tensor(m_source, dtype=DTYPE_COMPLEX, device=device)

    def set_device(self, device):
        super().set_device(device)
        self.r_source = self.r_source.to(device=device)
        self.m_source = self.m_source.to(device=device)

    def __repr__(self, verbose=False):
        out_str = " ------ {} -------".format(self.__name__)
        out_str += "\n - location (nm): {}".format(
            torch.round(self.r_source, decimals=1)
        )
        out_str += "\n - amplitude: {} ".format(self.p_source)

        return out_str

    def get_info(self):
        """Get info about field config. as dictionary"""
        return dict(
            name=self.__name__,
            r_source=self.r_source,
            m_source=self.m_source,
        )

    def get_x(self):
        """return x position of dipole"""
        return self.r_source.squeeze()[0]

    def get_y(self):
        """return y position of dipole"""
        return self.r_source.squeeze()[1]

    def get_z(self):
        """return z position of dipole"""
        return self.r_source.squeeze()[2]

    def get_efield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's electric field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: E-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        G_Em = environment.get_G_Em(r_probe, self.r_source.unsqueeze(0), wavelength)
        Em = torch.matmul(G_Em, self.m_source.unsqueeze(-1))[..., 0]
        return Em

    def get_hfield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's magnetic field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: H-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        G_Hm = environment.get_G_Hm(r_probe, self.r_source.unsqueeze(0), wavelength)
        Hm = torch.matmul(G_Hm, self.m_source.unsqueeze(-1))[..., 0]
        return Hm


# %%
if __name__ == "__main__":
    from torchgdm import env
    from torchgdm import tools
    import torchgdm as tg

    torch.set_printoptions(precision=3)
    # -- test plane wave
    wavelength = 700.0
    eps_env = 1.0
    env = env.dyads.EnvHomogeneous3D(env_material=eps_env)
    r_probe = torch.as_tensor(
        [[100, 0, 20], [0, 0, 100], [-100, 0, 200], [-500, 40, -200]], dtype=DTYPE_FLOAT
    )
    r_probe_xz = tools.geometry.coordinate_map_2d_square(2000, 51, 5, projection="xz")
    r_probe_yz = tools.geometry.coordinate_map_2d_square(2000, 51, 5, projection="yz")

    inangledeg = 60 * torch.pi / 180
    e_p = 0.5
    e_s = 0.2

    efield_xz = PlaneWave(e0p=e_p, e0s=e_s, inc_angle=inangledeg, inc_plane="xz")
    field_xz = efield_xz.get_field(r_probe_xz, wavelength=wavelength, environment=env)
    efield_yz = PlaneWave(e0p=e_p, e0s=e_s, inc_angle=inangledeg, inc_plane="yz")
    field_yz = efield_yz.get_field(r_probe_yz, wavelength=wavelength, environment=env)

    import matplotlib.pyplot as plt

    plt.figure(figsize=(8, 4))
    plt.subplot(221, aspect="equal")
    tg.visu.visu2d.vec2d.vectorfield(field_xz, whichfield="e")

    plt.subplot(223, aspect="equal")
    tg.visu.visu2d.vec2d.vectorfield(field_xz, whichfield="h")

    plt.subplot(222, aspect="equal")
    tg.visu.visu2d.vec2d.vectorfield(field_yz, whichfield="e")

    plt.subplot(224, aspect="equal")
    tg.visu.visu2d.vec2d.vectorfield(field_yz, whichfield="h")

    plt.show()



# encoding=utf-8
"""
internal tools for 2D visualizations
"""
# %%
import warnings

import torch

from torchgdm.tools.misc import to_np


def _get_axis_existing_or_new():
    import matplotlib.pyplot as plt

    if len(plt.get_fignums()) == 0:
        show = True
        ax = plt.subplot()
    else:
        show = False
        ax = plt.gca()
    return ax, show


def _get_closest_slice_level(levels, slice_level, limit_level=9999):
    import numpy as np

    slice_level_closest = np.unique(levels)[
        np.argmin(np.abs(np.unique(levels) - slice_level))
    ]
    if slice_level_closest != slice_level:
        if np.abs(slice_level) != limit_level:
            warnings.warn(
                "slice level at {}nm contains no meshpoints. Using closest level containing meshpoint at {}nm".format(
                    slice_level, slice_level_closest
                )
            )
    return slice_level_closest


def _automatic_projection(positions):
    import numpy as np

    pos_np = to_np(positions)

    if len(np.unique(pos_np[:, 2])) == 1:
        projection = "xy"
    elif len(np.unique(pos_np[:, 1])) == 1:
        projection = "xz"
    elif len(np.unique(pos_np[:, 0])) == 1:
        projection = "yz"
    else:
        # fallback
        warnings.warn("3D data. Falling back to XY projection...")
        projection = "xy"

    return projection


def _interpolate_to_grid(
    x,
    y,
    s,
    n_x=None,
    n_y=None,
    interpolation="nearest",
    fill_value="nan",
    return_grid=False,
    **kwargs
):
    """wrapper to scipy.interpolate.griddata

    List of X/Y/Z tuples to 2D x/y Grid with z as intensity value

    Parameters
    ----------
    x, y : np.array
        1D arrays containing x and y positions of the data

    s : np.array
        scalar values corresponding at each (x,y)-coordinate.

    n_x, n_y : int, optional
        number of points in X/Y direction for interpolation
        default: Number of unique occuring x/y elements

    interpolation : str, default: 'nearest'
        interpolation method for grid, passed to `scipy.griddata`

    **kwargs are passed to `scipy.interpolate.griddata`

    Returns
    -------
    map_ip : np.ndarray
        2d array with interpolated values in x/y plane

    extent : tuple
        x/y extent of array map as (x0, x1, y0, y1)

    """
    import numpy as np
    from scipy.interpolate import griddata

    if fill_value.lower == "nan":
        fill_value = np.nan

    # - determine limits and nr of steps for 2D grid interpolation
    x0, x1 = float(np.min(x)), float(np.max(x))
    y0, y1 = float(np.min(y)), float(np.max(y))
    extent = [x0, x1, y0, y1]

    if n_x is None:
        n_x = len(np.unique(x))
    if n_y is None:
        n_y = len(np.unique(y))

    # perform interpolation on regular grid
    grid_x, grid_y = np.meshgrid(np.linspace(x0, x1, n_x), np.linspace(y0, y1, n_y))
    map_ip = griddata(
        np.transpose([x, y]),
        s,
        (grid_x.T, grid_y.T),
        method=interpolation,
        fill_value=fill_value,
        **kwargs,
    )
    map_ip = np.flipud(map_ip.T)  # match convention: origin lower left
    if return_grid:
        return map_ip, extent, grid_x, grid_y
    else:
        return map_ip, extent


def _apply_projection(pos, projection):

    if projection.lower() == "xy":
        p = pos[:, [0, 1]]
        levels = pos[:, 2]
    elif projection.lower() == "yz":
        p = pos[:, [1, 2]]
        levels = pos[:, 0]
    elif projection.lower() == "xz":
        p = pos[:, [0, 2]]
        levels = pos[:, 1]
    else:
        raise ValueError("Invalid projection argument!")

    return p, levels


def _apply_projection_to_vectorfield(vec, pos, projection, complex_part=None):

    if complex_part is None:
        vec_scalar = vec
    elif complex_part.lower() == "real":
        vec_scalar = vec.real
    elif complex_part.lower() == "imag":
        vec_scalar = vec.imag
    else:
        raise ValueError(
            "Error: Unknown `complex_part` argument. Must be either 'real' or 'imag'."
        )

    v, _ = _apply_projection(vec_scalar, projection)
    p, levels = _apply_projection(pos, projection)

    return v, p, levels


def _optimize_alpha_value(
    pos2d_exp, alpha_step_limit=1e-3, alpha_init=0.0, time_limit=5.0
):
    """find best alpha_value that still returns an alphashape"""
    import logging
    import time
    import alphashape

    alpha_value = alpha_init
    step = 0.1
    t0 = time.time()

    # solver loop
    original_log_level = logging.getLogger().getEffectiveLevel()
    logging.disable(logging.WARN)
    while 1:
        _ga = alphashape.alphashape(pos2d_exp, alpha_value + step)
        print(alpha_value + step,_ga.geom_type)
        if _ga.geom_type not in ["Polygon", "MultiPolygon"]:
            step /= 2  # no solution: reduce search step
            if step < alpha_step_limit:
                break # step limit reached
            if time.time() - t0 > time_limit:
                break # time limit reached
        else:
            # solution found, update step
            alpha_value += step

    print(alpha_value,_ga.geom_type)
    logging.disable(original_log_level)
    return alpha_value


def _alpha_shape(points: torch.Tensor, alpha=1.0, only_outer=True):
    """Compute the alpha shape (concave hull) of a set of points.

    from : https://karobben.github.io/2022/03/07/Python/point_outline/

    Args:
        points (torch.Tensor): List of 2D points. shape (n,2) points.
        alpha (float, optional): alpha value. Defaults to 1.0.
        only_outer (bool, optional): whether to keep only the outer border or also inner edges. Defaults to True.

    Returns:
        torch.Tensor: set of (i,j) pairs representing edges of the alpha-shape. (i,j) are the indices in the points array.
    """
    from scipy.spatial import Delaunay

    assert points.shape[0] > 3, "Need at least four points"

    def add_edge(edges, i, j):
        """
        Add an edge between the i-th and j-th points,
        if not in the list already
        """
        if (i, j) in edges or (j, i) in edges:
            # already added
            assert (j, i) in edges, "Can't go twice over same directed edge right?"
            if only_outer:
                # if both neighboring triangles are in shape, it's not a boundary edge
                edges.remove((j, i))
            return
        edges.add((i, j))

    tri = Delaunay(points)

    edges = set()
    # Loop over triangles:
    # ia, ib, ic = indices of corner points of the triangle
    for ia, ib, ic in tri.simplices:
        pa = points[ia]
        pb = points[ib]
        pc = points[ic]
        # Computing radius of triangle circumcircle
        # www.mathalino.com/reviewer/derivation-of-formulas/derivation-of-formula-for-radius-of-circumcircle
        a = torch.sqrt((pa[0] - pb[0]) ** 2 + (pa[1] - pb[1]) ** 2)
        b = torch.sqrt((pb[0] - pc[0]) ** 2 + (pb[1] - pc[1]) ** 2)
        c = torch.sqrt((pc[0] - pa[0]) ** 2 + (pc[1] - pa[1]) ** 2)
        s = (a + b + c) / 2.0
        area = torch.sqrt(s * (s - a) * (s - b) * (s - c))
        circum_r = a * b * c / (4.0 * area)
        if circum_r < alpha:
            add_edge(edges, ia, ib)
            add_edge(edges, ib, ic)
            add_edge(edges, ic, ia)
    return edges



# -*- coding: utf-8 -*-
"""
scattering / extinction / absorption cross sections    
"""
# %%
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.tools.misc import get_closest_wavelength
from torchgdm.tools.batch import get_p_m_moments, get_e_hfields
from torchgdm.tools.batch import get_e0, get_h0
from torchgdm.tools.batch import get_p_m_radiative_correction_prefactors


def ecs(
    sim,
    wavelength: float,
    illumination_index: int = None,  # None: batch all illumination fields
    normalize_to_local_e0: bool = False,
    **kwargs,
):
    """total extinction cross section

    further kwargs are ignored

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): in nm
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        normalize_to_local_e0 (bool, optional): Whether to normalize to local illumination amplitude. Can be useful to get an approximate cross-section for non-plane wave illuminations. Defaults to False.

    Returns:
        dict: dictionaty containing the extinction cross section
    """

    from torchgdm.tools.misc import test_illumination_field_is_plane_wave

    # =============================================================================
    #     Exception handling
    # =============================================================================

    wavelength = get_closest_wavelength(sim, wavelength)

    test_illumination_field_is_plane_wave(
        sim,
        illumination_index,
        message="Extinction cross sections are only defined for plane wave illuminations! ",
    )

    # =============================================================================
    #     main calculation
    # =============================================================================

    # source positions and dipole moments
    p, m, r_p, r_m = get_p_m_moments(sim, wavelength, illumination_index)

    k0 = 2.0 * torch.pi / wavelength

    # environment ref. index at each source
    eps_env_at_p = sim.environment.get_environment_permittivity_scalar(
        wavelength, r_probe=r_p
    )
    eps_env_at_m = sim.environment.get_environment_permittivity_scalar(
        wavelength, r_probe=r_m
    )
    n_env_at_p = (eps_env_at_p**0.5).real
    n_env_at_m = (eps_env_at_m**0.5).real

    # optionally normalize each contribution to the local field amplitude
    e0_at_p = get_e0(sim, wavelength, r_p, illumination_index)
    h0_at_m = get_h0(sim, wavelength, r_m, illumination_index)
    if normalize_to_local_e0:
        # illumination fields
        e0_at_m = get_e0(sim, wavelength, r_m, illumination_index)
        # intensities
        I0_at_p = torch.sum(torch.abs(e0_at_p) ** 2, dim=-1)
        I0_at_m = torch.sum(torch.abs(e0_at_m) ** 2, dim=-1)
    else:
        I0_at_p = torch.ones(p.shape[:-1], device=p.device)
        I0_at_m = torch.ones(m.shape[:-1], device=m.device)

    ext_factor_p = (4 * torch.pi * k0 / (n_env_at_p * I0_at_p)).real.unsqueeze(-1)
    ext_factor_m = (4 * torch.pi * k0 / (n_env_at_m**2 * I0_at_m)).real.unsqueeze(-1)

    cs_ext_p = (ext_factor_p * torch.multiply(torch.conj(e0_at_p), p)).imag.sum((1, 2))
    cs_ext_m = (ext_factor_m * torch.multiply(torch.conj(h0_at_m), m)).imag.sum((1, 2))
    cs_ext = cs_ext_p + cs_ext_m
    return dict(ecs=cs_ext.real)


def acs(
    sim,
    wavelength: float,
    illumination_index: int = None,  # None: batch all illumination fields
    normalize_to_local_e0: bool = False,
    radiative_correction: bool = True,
    **kwargs,
):
    """total absorption cross section

    further kwargs are ignored

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): in nm
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

    Returns:
        dict: dictionaty containing the absorption cross section
    """
    # c.f. [Chaumet & Rahmani. JQSRT 110, 22–29 (2009), doi: 10.1016/j.jqsrt.2008.09.004]
    from torchgdm.tools.misc import test_illumination_field_is_plane_wave

    # =============================================================================
    #     Exception handling
    # =============================================================================

    wavelength = get_closest_wavelength(sim, wavelength)

    test_illumination_field_is_plane_wave(
        sim,
        illumination_index,
        message="Absorption cross sections are only defined for plane wave illuminations! ",
    )

    # =============================================================================
    #     main calculation
    # =============================================================================

    # source positions and dipole moments
    p, m, r_p, r_m = get_p_m_moments(sim, wavelength, illumination_index)
    e_at_p, h_at_m, r_p, r_m = get_e_hfields(sim, wavelength, illumination_index)

    k0 = 2.0 * torch.pi / wavelength

    # environment ref. index at each source
    eps_env_at_p = sim.environment.get_environment_permittivity_scalar(
        wavelength, r_probe=r_p
    )
    eps_env_at_m = sim.environment.get_environment_permittivity_scalar(
        wavelength, r_probe=r_m
    )
    n_env_at_p = (eps_env_at_p**0.5).real
    n_env_at_m = (eps_env_at_m**0.5).real

    # optionally normalize each contribution to the local field amplitude
    if normalize_to_local_e0:
        # illumination fields
        e0_at_p = get_e0(sim, wavelength, r_p, illumination_index)
        e0_at_m = get_e0(sim, wavelength, r_m, illumination_index)
        # intensities
        I0_at_p = torch.sum(torch.abs(e0_at_p) ** 2, dim=-1)
        I0_at_m = torch.sum(torch.abs(e0_at_m) ** 2, dim=-1)
    else:
        I0_at_p = torch.ones(p.shape[:-1], device=p.device)
        I0_at_m = torch.ones(m.shape[:-1], device=m.device)

    # --- absorption
    abs_factor_p = (4 * torch.pi * k0 / (n_env_at_p * I0_at_p)).real.unsqueeze(-1)
    abs_factor_m = (4 * torch.pi * k0 / (n_env_at_m * I0_at_m)).real.unsqueeze(-1)

    cs_abs_p = (abs_factor_p * torch.multiply(p, torch.conj(e_at_p))).imag.sum((1, 2))
    cs_abs_m = (abs_factor_m * torch.multiply(m, torch.conj(h_at_m))).imag.sum((1, 2))

    # radiative correction:
    f_rad_p, f_rad_m = get_p_m_radiative_correction_prefactors(sim, wavelength)

    rad_corr_p = (
        abs_factor_p * (torch.multiply(p * f_rad_p.unsqueeze(0), torch.conj(p)))
    ).sum((1, 2))

    rad_corr_m = (
        abs_factor_m * (torch.multiply(m * f_rad_m.unsqueeze(0), torch.conj(m)))
    ).sum((1, 2))

    cs_abs_p = cs_abs_p - rad_corr_p
    cs_abs_m = cs_abs_m - rad_corr_m

    cs_abs = (cs_abs_p + cs_abs_m).real

    return dict(acs=cs_abs.real)


def scs(
    sim,
    wavelength: float,
    illumination_index: int = None,  # None: batch all illumination fields
    normalize_to_local_e0: bool = False,
    radiative_correction: bool = True,
    **kwargs,
):
    """total scattering cross section

    This function also returns extinction and absorption cross-sections, which are calculated as byproducts

    further kwargs are ignored

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): in nm
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        normalize_to_local_e0 (bool, optional): Whether to normalize to local illumination amplitude. Can be useful to get an approximate cross-section for non-plane wave illuminations. Defaults to False.
        radiative_correction (bool, optional): Whether to add radiative correction term. Defaults to True.

    Returns:
        dict: dictionaty containing the scattering, absorption and extinction cross sections
    """
    from torchgdm.tools.misc import test_illumination_field_is_plane_wave

    # =============================================================================
    #     Exception handling
    # =============================================================================

    wavelength = get_closest_wavelength(sim, wavelength)

    test_illumination_field_is_plane_wave(
        sim,
        illumination_index,
        message="Scattering cross sections are only defined for plane wave illuminations! ",
    )

    # =============================================================================
    #     main calculation
    # =============================================================================

    # --- general pre-calculation
    # source positions and dipole moments
    p, m, r_p, r_m = get_p_m_moments(sim, wavelength, illumination_index)
    e_at_p, h_at_m, r_p, r_m = get_e_hfields(sim, wavelength, illumination_index)

    k0 = 2.0 * torch.pi / wavelength

    # environment ref. index at each source
    eps_env_at_p = sim.environment.get_environment_permittivity_scalar(
        wavelength, r_probe=r_p
    )
    eps_env_at_m = sim.environment.get_environment_permittivity_scalar(
        wavelength, r_probe=r_m
    )
    n_env_at_p = (eps_env_at_p**0.5).real
    n_env_at_m = (eps_env_at_m**0.5).real

    # illumination field
    e0_at_p = get_e0(sim, wavelength, r_p, illumination_index)
    e0_at_m = get_e0(sim, wavelength, r_m, illumination_index)
    h0_at_m = get_h0(sim, wavelength, r_m, illumination_index)

    # optionally normalize each contribution to the local field amplitude
    if normalize_to_local_e0:
        # intensities
        I0_at_p = torch.sum(torch.abs(e0_at_p) ** 2, dim=-1)
        I0_at_m = torch.sum(torch.abs(e0_at_m) ** 2, dim=-1)
    else:
        I0_at_p = torch.ones(p.shape[:-1], device=p.device)
        I0_at_m = torch.ones(m.shape[:-1], device=m.device)

    # --- absorption
    abs_factor_p = (4 * torch.pi * k0 / (n_env_at_p * I0_at_p)).real.unsqueeze(-1)
    abs_factor_m = (4 * torch.pi * k0 / (n_env_at_m * I0_at_m)).real.unsqueeze(-1)

    cs_abs_p = (abs_factor_p * torch.multiply(p, torch.conj(e_at_p))).imag.sum((1, 2))
    cs_abs_m = (abs_factor_m * torch.multiply(m, torch.conj(h_at_m))).imag.sum((1, 2))

    # radiative correction:
    f_rad_p, f_rad_m = get_p_m_radiative_correction_prefactors(sim, wavelength)

    rad_corr_p = (
        abs_factor_p * (torch.multiply(p * f_rad_p.unsqueeze(0), torch.conj(p)))
    ).sum((1, 2))

    rad_corr_m = (
        abs_factor_m * (torch.multiply(m * f_rad_m.unsqueeze(0), torch.conj(m)))
    ).sum((1, 2))

    cs_abs_p = cs_abs_p - rad_corr_p
    cs_abs_m = cs_abs_m - rad_corr_m

    cs_abs = (cs_abs_p + cs_abs_m).real

    # --- extinction
    ext_factor_p = (4 * torch.pi * k0 / (n_env_at_p * I0_at_p)).real.unsqueeze(-1)
    ext_factor_m = (4 * torch.pi * k0 / (n_env_at_m * I0_at_m)).real.unsqueeze(-1)

    cs_ext_p = (ext_factor_p * torch.multiply(torch.conj(e0_at_p), p)).imag.sum((1, 2))
    cs_ext_m = (ext_factor_m * torch.multiply(torch.conj(h0_at_m), m)).imag.sum((1, 2))
    cs_ext = (cs_ext_p + cs_ext_m).real

    # --- scattering
    cs_scat = cs_ext - cs_abs

    return dict(scs=cs_scat, ecs=cs_ext, acs=cs_abs)


def total(
    sim,
    wavelength: float,
    illumination_index: int = None,  # None: batch all illumination fields
    normalize_to_local_e0: bool = False,
    radiative_correction: bool = True,
    **kwargs,
):
    """compute total cross sections: extinction, absorption and scattering

    This is a wrapper to :func:`scs`.

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): in nm
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        normalize_to_local_e0 (bool, optional): Whether to normalize to local illumination amplitude. Can be useful to get an approximate cross-section for non-plane wave illuminations. Defaults to False.
        radiative_correction (bool, optional): Whether to add radiative correction term. Defaults to True.

    Returns:
        dict: dictionaty containing the scattering, absorption and extinction cross sections
    """

    return scs(
        sim=sim,
        wavelength=wavelength,
        illumination_index=illumination_index,
        normalize_to_local_e0=normalize_to_local_e0,
        radiative_correction=radiative_correction,
        **kwargs,
    )


# --- testing
if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""3D visualization tools based on pyvista

requirements:
 - numpy
 - pyvista
 
for use in jupyter:
 - trame
 - trame-vtk
 - trame-vuetify
 - ipywidgets
 
install with:

     pip install 'jupyterlab>=3' ipywidgets 'pyvista[all,trame]'
 
"""
from . import geo3d
from . import vec3d

from .geo3d import structure
from .vec3d import vectorfield, vectorfield_inside



# -*- coding: utf-8 -*-
"""volume discretization for torchgdm
"""
from .pola import StructDiscretized2D, StructDiscretizedSquare2D
from . import pola
from . import geometries

from .geometries import discretizer_square

from .geometries import square
from .geometries import rectangle
from .geometries import circle
from .geometries import split_ring
from .geometries import triangle_equilateral



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestFieldCalculations2D(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing fields postprocessing with discretized 2D-simulation...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        eps_env = 1.0
        mat_env = tg.materials.MatConstant(eps_env)
        env = tg.env.freespace_2d.EnvHomogeneous2D(env_material=mat_env)

        # - illumination field(s)
        self.wavelengths = torch.tensor([500.0])
        e_inc_list = [
            tg.env.freespace_2d.PlaneWave(
                e0p=0.5, e0s=0.7, inc_angle=torch.pi
            ),  # linear-polarization, incidence from top
        ]

        # - structure
        step = 15.0
        mat_struct = tg.materials.MatConstant(eps=7.0)
        struct2d = tg.struct.surface_2d.StructDiscretizedSquare2D(
            tg.struct.surface_2d.geometries.rectangle(l=5, h=3), step, mat_struct
        )

        self.sim = tg.simulation.Simulation(
            structures=[struct2d],
            environment=env,
            illumination_fields=e_inc_list,
            wavelengths=self.wavelengths,
        )

    def test_internal_fields(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(verbose=False, progress_bar=False, calc_missing=True)

            # calculate internal fields
            self.r_probe = self.sim.get_all_positions()
            nf_res = tg.postproc.fields.nf(
                self.sim,
                self.wavelengths[0],
                self.r_probe,
                interpolation_step_range=0.1,
                source_distance_steps=0.1,
                progress_bar=False,
            )
            nf_postproc = nf_res["tot"]
            nf_inside = self.sim.fields_inside[float(self.wavelengths[0])]

            torch.testing.assert_close(
                nf_postproc.efield, nf_inside.efield, rtol=1e-3, atol=1e-3
            )
            torch.testing.assert_close(
                nf_postproc.hfield, nf_inside.hfield, rtol=1e-3, atol=1e-3
            )

            if self.verbose:
                print(
                    "  - {}: 2D-sim internal field consistency test passed.".format(
                        device
                    )
                )

    def test_nearfield(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(verbose=False, progress_bar=False)

            # calculate nearfield
            self.r_probe = torch.as_tensor(
                [[300.0, 00.0, 105.0], [12.0, 0.0, -50.0]], device=device
            )
            nf_res = tg.postproc.fields._nearfield(
                self.sim,
                self.wavelengths[0],
                self.r_probe,
            )
            nf = torch.stack([nf_res["e_sca"], nf_res["e_tot"], nf_res["h_tot"]])
            # print(nf)

            # compare to pre-calculated reference
            nf_truth = torch.as_tensor(
                [
                    [
                        [
                            [
                                -0.0123919 - 0.0022812j,
                                0.1676146 - 0.1816293j,
                                -0.0049116 + 0.0124582j,
                            ],
                            [
                                -0.1053862 + 0.0654174j,
                                -0.3760716 + 0.3414206j,
                                -0.0329867 + 0.0019590j,
                            ],
                        ]
                    ],
                    [
                        [
                            [
                                0.1119532 - 0.4865728j,
                                0.3416977 - 0.8596374j,
                                -0.0049116 + 0.0124581j,
                            ],
                            [
                                0.2991223 + 0.3593100j,
                                0.1902403 + 0.7528703j,
                                -0.0329866 + 0.0019590j,
                            ],
                        ]
                    ],
                    [
                        [
                            [
                                0.1220035 - 0.6340325j,
                                -0.1247076 + 0.4713390j,
                                0.1848267 - 0.1560034j,
                            ],
                            [
                                0.0178721 + 0.5906836j,
                                -0.3645644 - 0.4154846j,
                                -0.0858047 + 0.0229988j,
                            ],
                        ]
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            torch.testing.assert_close(nf, nf_truth, rtol=1e-3, atol=1e-3)

            if self.verbose:
                print("  - {}: 3D-sim nearfield test passed.".format(device))

    def test_field_gradients(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(verbose=False, progress_bar=False)

            # calculate field gradients
            self.r_probe = torch.as_tensor(
                [[0.0, 0.0, 105.0], [12.0, 0.0, 200.0]],
                dtype=tg.constants.DTYPE_FLOAT,
                device=device,
            )
            delta = 0.1
            field = "e_tot"
            fgrads = tg.postproc.fields.field_gradient(
                self.sim,
                self.wavelengths[0],
                self.r_probe,
                delta=delta,
                whichfield=field,
            )
            grad = torch.stack([fgrads["dfdx"], fgrads["dfdy"], fgrads["dfdz"]])
            # print(grad)

            # compare to pre-calculated reference
            grad_truth = torch.as_tensor(
                [
                    [
                        [
                            [
                                -3.7252903e-08 + 0.0000000e00j,
                                1.4901161e-07 - 1.4901161e-07j,
                                2.1063897e-03 - 1.5938655e-04j,
                            ],
                            [
                                7.2568655e-05 - 1.2069941e-05j,
                                2.1457672e-04 - 1.8179417e-04j,
                                3.8543949e-04 + 2.3508561e-04j,
                            ],
                        ]
                    ],
                    [
                        [
                            [
                                0.0000000e00 + 0.0000000e00j,
                                0.0000000e00 + 0.0000000e00j,
                                0.0000000e00 + 0.0000000e00j,
                            ],
                            [
                                0.0000000e00 + 0.0000000e00j,
                                0.0000000e00 + 0.0000000e00j,
                                0.0000000e00 + 0.0000000e00j,
                            ],
                        ]
                    ],
                    [
                        [
                            [
                                -5.2526221e-03 - 2.2722781e-03j,
                                -9.8736584e-03 - 8.6054206e-03j,
                                2.9604230e-09 - 1.3007728e-09j,
                            ],
                            [
                                -3.3956766e-03 + 4.4409931e-03j,
                                -1.9621849e-03 + 4.3258071e-03j,
                                -7.2531402e-05 + 1.2074597e-05j,
                            ],
                        ]
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            torch.testing.assert_close(grad, grad_truth, rtol=1e-3, atol=1e-3)

            if self.verbose:
                print("  - {}: 3D-sim field gradients test passed.".format(device))


class TestFieldCalculations3D(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing fields postprocessing with discretized 3D-simulation...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        eps_env = 1.0
        mat_env = tg.materials.MatConstant(eps_env)
        env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

        # - illumination field(s)
        self.wavelengths = torch.tensor([500.0])
        e_inc_list = [
            tg.env.freespace_3d.PlaneWave(
                e0p=1.0, e0s=0.0, inc_angle=torch.pi
            ),  # X-polarization, incidence from top
        ]

        # - structure
        step = 15.0
        mat_struct = tg.materials.MatConstant(eps=9.0)
        struct1 = tg.struct.volume.StructDiscretizedCubic3D(
            tg.struct.volume.geometries.cuboid(l=3, w=4, h=2),
            step,
            mat_struct,
            radiative_correction=False,
        )

        self.sim = tg.simulation.Simulation(
            structures=[struct1],
            environment=env,
            illumination_fields=e_inc_list,
            wavelengths=self.wavelengths,
        )

    def test_internal_fields(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(verbose=False, progress_bar=False, calc_missing=True)

            # calculate internal fields
            self.r_probe = self.sim.get_all_positions()
            nf_res = tg.postproc.fields.nf(
                self.sim,
                self.wavelengths[0],
                self.r_probe,
                interpolation_step_range=0.1,
                source_distance_steps=0.1,
                progress_bar=False,
            )
            nf_postproc = nf_res["tot"]
            nf_inside = self.sim.fields_inside[float(self.wavelengths[0])]

            torch.testing.assert_close(
                nf_postproc.efield, nf_inside.efield, rtol=1e-3, atol=1e-3
            )
            torch.testing.assert_close(
                nf_postproc.hfield, nf_inside.hfield, rtol=1e-3, atol=1e-3
            )

            if self.verbose:
                print(
                    "  - {}: 3D-sim internal field consistency test passed.".format(
                        device
                    )
                )

    def test_nearfield(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(verbose=False, progress_bar=False)

            # calculate nearfield
            self.r_probe = torch.as_tensor(
                [[300.0, 40.0, 105.0], [12.0, 50.0, -50.0]], device=device
            )
            nf_res = tg.postproc.fields._nearfield(
                self.sim,
                self.wavelengths[0],
                self.r_probe,
            )
            nf = torch.stack([nf_res["e_sca"], nf_res["e_tot"], nf_res["h_tot"]])
            # print(nf)

            # compare to pre-calculated reference
            nf_truth = torch.as_tensor(
                [
                    [
                        [
                            [
                                -3.4736176e-03 + 1.8434002e-03j,
                                1.4439697e-04 + 8.7944919e-04j,
                                3.6897586e-04 + 2.4145723e-03j,
                            ],
                            [
                                -2.1888228e-02 + 2.1395750e-02j,
                                8.2111722e-03 - 9.2640589e-04j,
                                -1.5400744e-02 + 1.9592703e-03j,
                            ],
                        ]
                    ],
                    [
                        [
                            [
                                2.4521643e-01 - 9.6673971e-01j,
                                1.4439697e-04 + 8.7944919e-04j,
                                3.6899760e-04 + 2.4144875e-03j,
                            ],
                            [
                                7.8712869e-01 + 6.0918105e-01j,
                                8.2111722e-03 - 9.2640589e-04j,
                                -1.5400673e-02 + 1.9593218e-03j,
                            ],
                        ]
                    ],
                    [
                        [
                            [
                                -5.3811527e-07 - 5.9395365e-06j,
                                -2.5016230e-01 + 9.6671307e-01j,
                                5.3901912e-04 + 6.8496628e-04j,
                            ],
                            [
                                -4.0306677e-05 + 1.0966815e-04j,
                                -8.0708373e-01 - 6.2726992e-01j,
                                2.1683499e-03 - 2.3433529e-02j,
                            ],
                        ]
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            torch.testing.assert_close(nf, nf_truth, rtol=1e-3, atol=1e-3)

            if self.verbose:
                print("  - {}: 3D-sim nearfield test passed.".format(device))

    def test_field_gradients(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(verbose=False, progress_bar=False)

            # calculate field gradients
            self.r_probe = torch.as_tensor(
                [[0.0, 0.0, 105.0], [12.0, 50.0, 200.0]],
                dtype=tg.constants.DTYPE_FLOAT,
                device=device,
            )
            delta = 0.1
            field = "e_tot"
            fgrads = tg.postproc.fields.field_gradient(
                self.sim,
                self.wavelengths[0],
                self.r_probe,
                delta=delta,
                whichfield=field,
            )
            grad = torch.stack([fgrads["dfdx"], fgrads["dfdy"], fgrads["dfdz"]])
            # print(grad)

            # compare to pre-calculated reference
            grad_truth = torch.as_tensor(
                [
                    [
                        [
                            [
                                0.0000000e00 + 0.0000000e00j,
                                -5.7825473e-05 + 9.4272946e-06j,
                                8.1359642e-04 - 1.3028191e-04j,
                            ],
                            [
                                1.1026859e-05 - 4.7683716e-06j,
                                1.9397121e-05 + 5.9410741e-06j,
                                8.6341170e-05 + 2.6219059e-05j,
                            ],
                        ]
                    ],
                    [
                        [
                            [
                                -2.2500753e-05 + 1.3709068e-05j,
                                2.9103830e-10 + 2.0918378e-10j,
                                2.9103830e-10 - 2.9103830e-10j,
                            ],
                            [
                                -5.9604645e-07 - 2.9206276e-05j,
                                4.5253546e-06 + 1.7492493e-06j,
                                -4.7066715e-06 + 3.4109689e-07j,
                            ],
                        ]
                    ],
                    [
                        [
                            [
                                -1.1811852e-02 - 3.3086538e-03j,
                                4.3655746e-10 - 5.4569682e-11j,
                                -1.3551205e-09 - 1.0508572e-09j,
                            ],
                            [
                                -7.3888898e-03 + 1.0038614e-02j,
                                -4.7260255e-06 + 3.6117854e-07j,
                                -1.5694532e-05 + 3.2107346e-06j,
                            ],
                        ]
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            torch.testing.assert_close(grad, grad_truth, rtol=1e-3, atol=1e-3)

            if self.verbose:
                print("  - {}: 3D-sim field gradients test passed.".format(device))


# %%

# %%

# %%
if __name__ == "__main__":
    print("testing fields postprocessing...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""package for various sorts of tools for torchgdm
"""
from . import geometry
from . import batch
from . import misc
from . import interp
from . import special
from . import mie



# -*- coding: utf-8 -*-
"""volume discretization structure class
"""
import warnings

import torch

from torchgdm.constants import (
    DTYPE_FLOAT,
    DTYPE_COMPLEX,
    COLORS_DEFAULT,
)
from torchgdm.tools.misc import get_default_device
from torchgdm.struct.base_classes import StructBase
from torchgdm.tools.geometry import get_step_from_geometry
from torchgdm.tools.geometry import sample_random_circular
from torchgdm.tools.geometry import rotation_y
from torchgdm.tools.misc import ptp


class StructDiscretized2D(StructBase):
    """base class 2D surface discretized structure with infinite axis

    Defines the 2D surface discretization, its polarizabilities and self-terms
    """

    __name__ = "2D discretized structure"

    def __init__(
        self,
        positions: torch.Tensor,
        materials,
        step=None,
        radiative_correction: bool = False,
        device: torch.device = None,
        **kwargs
    ):
        """2D discretized structure

        Args:
            positions (torch.Tensor): meshpoint positions (3D, but all y values must be zero)
            materials (material class, or list of them): material of structure. If list, define the material of each meshpoint.
            step (float, optional): nominal stepsize. If not provided, infer automatically from geometry. Defaults to None.
            radiative_correction (bool, optional): Whether to add raditative correction for cross section calculations. Defaults to False.
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).

        Raises:
            ValueError: No mesh points at y=0, Invalid material config
        """
        super().__init__(device=device, **kwargs)
        self.mesh = "2D"
        self.n_dim = 2

        self.positions = torch.as_tensor(
            positions, dtype=DTYPE_FLOAT, device=self.device
        )
        if torch.count_nonzero(self.positions[..., 1]) > 0:
            warnings.warn("2D structure. Remove all positions with y!=0.")
            self.positions = self.positions[self.positions[..., 1] != 0]
            if len(self.positions) == 0:
                raise ValueError("No mesh positions at y=0. Please check geometry.")

        self.r0 = self.get_center_of_mass()  # center of gravity

        if step is None:
            step_scalar = get_step_from_geometry(self.positions)
        else:
            step_scalar = step

        # step for every meshcell, for consistency with other struct classes
        self.step = step_scalar * torch.ones(
            len(self.positions), dtype=DTYPE_FLOAT, device=self.device
        )

        self.radiative_correction = radiative_correction
        self.mesh_normalization_factor = torch.as_tensor(
            1, dtype=DTYPE_FLOAT, device=self.device
        )  # square mesh

        # material of each meshpoint
        if hasattr(materials, "__iter__"):
            if len(materials) != len(self.positions):
                raise ValueError(
                    "Either a global material needs to be given or "
                    + "each meshpoint needs a defined material. "
                    + "But meshpoint list and materials list are of different lengths."
                )
            self.materials = materials
        else:
            self.materials = [materials for i in self.positions]

        self.zeros = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )

        # discretized, made from natural material: only electric response
        self.evaluation_terms = ["E"]  # possible terms 'E' and 'H'

    def __repr__(self, verbose=False):
        """description about structure"""
        out_str = ""
        out_str += "------ discretized 2D nano-object -------"
        out_str += "\n" + " mesh type:              {}".format(self.mesh)
        out_str += "\n" + " nr. of meshpoints:      {}".format(len(self.positions))
        out_str += "\n" + " nominal stepsizes (nm): {}".format(
            [float(f) for f in torch.unique(self.step)]
        )
        out_str += "\n" + " material:               {}".format(
            [m.__name__ for m in set(self.materials)]
        )
        bnds = ptp(self.positions, dim=0)
        out_str += "\n" + " size & position (Y-axis is infinite):"
        out_str += "\n" + "     X-extension          :   {:.1f} (nm)".format(bnds[0])
        out_str += "\n" + "     Z-extension          :   {:.1f} (nm)".format(bnds[2])
        out_str += "\n" + "     center of mass (x,z) : ({:.1f}, {:.1f})".format(
            self.r0[0], self.r0[2]
        )

        return out_str

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)

        self.zeros = self.zeros.to(device)
        self.step = self.step.to(device)
        self.r0 = self.r0.to(device)
        self.mesh_normalization_factor = self.mesh_normalization_factor.to(device)

        for mat in self.materials:
            mat.set_device(device)

    # --- self-terms
    def get_selfterm_pE(self, wavelength: float, environment) -> torch.Tensor:
        """return list of 'EE' self-term tensors (3x3) at each meshpoint

        Args:
            wavelength (float): in nm
            environment (environment class): 2D environement class

        Returns:
            torch.Tensor: pE self term tensor
        """
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=self.positions
        )
        # cast env permittivity to real, because hankel only support real args
        eps_env = torch.as_tensor(eps_env, dtype=DTYPE_COMPLEX, device=self.device).real

        k0 = 2 * torch.pi / wavelength
        k02 = k0**2

        if self.mesh_normalization_factor == 0:
            norm_xz = 0
            norm_y = 0
        else:
            from torchgdm.tools.special import H1n

            S = self.step**2
            k0_y = environment.get_k0_y(wavelength)

            kr2 = torch.as_tensor(
                eps_env * k02 - k0_y**2, dtype=DTYPE_FLOAT, device=self.device
            )
            kr = torch.sqrt(kr2)

            h11 = H1n(1, kr * self.step / torch.pi**0.5)
            norm01 = self.step / torch.pi**0.5 * h11 / kr + 2j / (torch.pi * kr**2)

            norm_xz_nonrad = -1 * self.mesh_normalization_factor / (2.0 * S * eps_env)
            norm_xz_rad = 1j * torch.pi * (2 * k02 - kr2 / eps_env) * norm01 / (4 * S)

            norm_y_nonrad = 0
            norm_y_rad = 1j * torch.pi * (k02 - k0_y**2 / eps_env) * norm01 / (2 * S)

            norm_xz = 4.0 * torch.pi * (norm_xz_nonrad + norm_xz_rad)
            norm_y = 4.0 * torch.pi * (norm_y_nonrad + norm_y_rad)

        self_terms_pE = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        self_terms_pE[:, 0, 0] = norm_xz
        self_terms_pE[:, 1, 1] = norm_y
        self_terms_pE[:, 2, 2] = norm_xz

        return self_terms_pE

    # --- polarizabilities
    def get_polarizability_pE(self, wavelength: float, environment) -> torch.Tensor:
        """return list of EE polarizability tensors (3x3) at each meshpoint

        Args:
            wavelength (float): in nm
            environment (environment class): 2D environement class

        Returns:
            torch.Tensor: pE polarizability tensor
        """
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=self.positions
        )
        eps_env_tensor = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        eps_env_tensor[:, 0, 0] = eps_env
        eps_env_tensor[:, 1, 1] = eps_env
        eps_env_tensor[:, 2, 2] = eps_env

        eps_geo = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        for i, mat in enumerate(self.materials):
            eps_geo[i] = mat.get_epsilon(wavelength)

        S_cell_norm = self.step**2 / float(self.mesh_normalization_factor)

        # --- polarizability
        alpha_pE = (
            (eps_geo - eps_env_tensor)
            * S_cell_norm.unsqueeze(1).unsqueeze(1)
            / (4.0 * torch.pi)
        )

        return alpha_pE

    # - radiative correction for cross section calc. - 2D case
    def get_radiative_correction_prefactor_p(self, wavelength: float, environment):
        """return electric dipole radiative correction prefactor vectors (3) at each meshpoint"""
        if self.radiative_correction:
            k0 = 2 * torch.pi / wavelength

            pf_vec = torch.as_tensor(
                [1.0, 2.0, 1.0], device=self.device, dtype=DTYPE_COMPLEX
            )
            pf_vec = pf_vec * torch.pi / 2 * k0**2

            return torch.ones(
                (len(self.positions), 3), device=self.device, dtype=DTYPE_COMPLEX
            ) * pf_vec.unsqueeze(0)
        else:
            return torch.zeros(
                (len(self.positions), 3), device=self.device, dtype=DTYPE_COMPLEX
            )

    def get_radiative_correction_prefactor_m(self, wavelength: float, environment):
        """return magnetic dipole radiative correction prefactor vectors (3) at each meshpoint"""
        if self.radiative_correction:
            k0 = 2 * torch.pi / wavelength
            n_env = (
                environment.get_environment_permittivity_scalar(
                    wavelength, self.positions
                )
                ** 0.5
            )
            pf_vec = torch.as_tensor(
                [1.0, 2.0, 1.0], device=self.device, dtype=DTYPE_COMPLEX
            )
            pf_vec = pf_vec * torch.pi / 2 * k0**2

            return (
                torch.ones(
                    (len(self.positions), 3), device=self.device, dtype=DTYPE_COMPLEX
                )
                * pf_vec.unsqueeze(0)
                * n_env.unsqueeze(1) ** 2
            )
        else:
            return torch.zeros(
                (len(self.positions), 3), device=self.device, dtype=DTYPE_COMPLEX
            )

    def convert_to_effective_polarizability_pair(
        self, environment, wavelengths, test_accuracy=False, **kwargs
    ):
        from torchgdm.struct import StructEffPola2D
        from torchgdm.struct.volume.pola import test_effective_polarizability_accuracy

        warnings.warn(
            "2D effective polarizabilities only implemented for illumination incidence in XZ plane!"
        )

        alpha = extract_eff_pola_via_propagation(
            struct=self, environment=environment, wavelengths=wavelengths, **kwargs
        )
        struct_aeff = StructEffPola2D(
            positions=torch.stack([self.r0]),
            alpha_dicts=[alpha],
            environment=environment,
            device=self.device,
        )

        if test_accuracy:
            test_effective_polarizability_accuracy(
                struct_aeff, self, test_yz_incidence=False
            )

        return struct_aeff

    # --- plotting
    def plot(self, **kwargs):
        """plot the structure in XZ plane (2D)"""
        from torchgdm.visu import visu2d

        kwargs["projection"] = "xz"
        im = visu2d.geo2d._plot_structure_discretized(self, **kwargs)
        return im

    def plot_contour(self, **kwargs):
        """plot the structure contour in XZ plane (2D)"""
        from torchgdm.visu import visu2d

        kwargs["projection"] = "xz"
        im = visu2d.geo2d._plot_contour_discretized(self, **kwargs)
        return im

    def plot3d(self, **kwargs):
        """plot the structure in 3D"""
        from torchgdm.visu import visu3d

        warnings.warn(
            "Visualizing a 2D model in 3D will show only meshpoints in the 'XZ' plane."
        )

        return visu3d.geo3d._plot_structure_discretized(self, **kwargs)

    # --- geometry operations
    def rotate(
        self,
        alpha: float,
        center: torch.Tensor = torch.as_tensor([0.0, 0.0, 0.0]),
        axis: str = "y",
    ):
        """rotate the structure

        Args:
            alpha (float): rotation angle (in rad)
            center (torch.Tensor, optional): center of rotation axis. Defaults to torch.as_tensor([0.0, 0.0, 0.0]).
            axis (str, optional): rotation axis. Defaults to "y".

        Raises:
            ValueError: only "y" axis supported in 2D

        Returns:
            :class:`StructDiscretized2D`: copy of structure with rotated geometry
        """
        _struct_rotated = self.copy()
        center = center.to(dtype=DTYPE_FLOAT, device=self.device)

        if axis.lower() == "y":
            rot = rotation_y(alpha, device=self.device)
        else:
            raise ValueError(
                "Only rotation axis 'y' supported in 2D (infinite axis).".format(axis)
            )

        if len(_struct_rotated.positions) > 1:
            _struct_rotated.positions = torch.matmul(
                _struct_rotated.positions - (center + self.r0), rot
            ) + (center + self.r0)
        else:
            warnings.warn("Single meshpoint found, ignore rotation.")

        return _struct_rotated


class StructDiscretizedSquare2D(StructDiscretized2D):
    """class for square surface discretized, infinitely long 2D structure

    Defines the square surface discretization, polarizabilities and self-terms
    """

    __name__ = "2D square lattice discretized structure class"

    def __init__(
        self,
        discretization_config,
        step,
        materials,
        device: torch.device = None,
        **kwargs
    ):
        """2D structure, discretized on a square lattice

        Infinite axis along Y

        Args:
            discretization_config (tuple): tuple of discretization condition function, and discretizer walk limits (as provided by the geometries generators)
            step (float, optional): nominal stepsize. If not provided, infer automatically from geometry. Defaults to None.
            materials (material class, or list of them): material of structure. If list, define the material of each meshpoint.
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).

        Raises:
            ValueError: No mesh points at y=0, Invalid material config
        """
        from torchgdm.struct.surface_2d import discretizer_square

        positions = discretizer_square(*discretization_config, step=step)

        super().__init__(positions, materials, step=step, device=device, **kwargs)

        self.mesh = "square (2D)"
        self.mesh_normalization_factor = torch.tensor(
            1.0, dtype=DTYPE_FLOAT, device=self.device
        )


def extract_eff_pola_via_propagation(
    struct,
    environment,
    wavelengths,
    n_probe=150,
    distance_probe=10.0,
    n_dipoles=None,
    distance_dipoles=5000,
    verbose=True,
    only_pE_mH=False,
    progress_bar=True,
    device=None,
):
    """Extract effective electric and magnetic dipole polarizability for volume discretized structure

    Extract the polarizability for the structure `struct` in a given `environement`
    at the specified `wavelengths`

    In this version, the effective dipole response for several illuminations is obtained
    via matching on a circumscribing sphere by solving of a first inverse problem via pseudoinverse.

    The second inverse problem of adjusting polarizability for different
    illuminations is also solved via pseudoinverse.

    By default, use 14 plane waves (different incidence directions and polarizations).
    alternative: illumination with `n_dipoles` point-dipole sources if `n_dipoles` is an integer > 0.


    Args:
        struct (class:`StructDiscretized3D`): the discretized structure to extract the model for.
        environment (environment class): 3D environement class.
        wavelengths (torch.Tensor): list of wavelengths to extract the model at. in nm.
        n_probe (int, optional): number of probe positions on enclosing sphere. Defaults to 100.
        distance_probe (float, optional): additional distance to enclosing sphere in units of discretization step. Defaults to 2.0.
        n_dipoles (int, optional): If given, use `n` dipole sources as illumination instead of plane waves. Defaults to None.
        distance_dipoles (int, optional): if using dipoles, specify their distance to the center of gravity. Defaults to 5000.
        verbose (bool, optional): whether to sprint progess info. Defaults to True.
        only_pE_mH (bool, optional): whether to extract only a p/m model (True) or a full (6x6) polarizability (False). Defaults to True.
        progress_bar (bool, optional): whether to show a progress bar. Defaults to True.
        device (str, optional). Defaults to None, in which case the structure's device is used.

    Returns:
        dict: contains all informations about the effective dipole polarizability model
    """
    from torchgdm.env.freespace_2d.inc_fields import ElectricLineDipole, PlaneWave
    from torchgdm.simulation import Simulation
    from torchgdm.postproc.fields import nf
    from torchgdm.tools.geometry import get_enclosing_sphere_radius
    from torchgdm.tools.misc import tqdm
    from torchgdm.struct.volume.pola import _test_residual_effective_polarizability

    if verbose:
        import time
    _struct = struct.copy()

    if device is None:
        device = _struct.device
    else:
        struct.set_device(device)

    # convert single wavelength to list
    wavelengths = torch.as_tensor(wavelengths, dtype=DTYPE_FLOAT, device=device)
    wavelengths = torch.atleast_1d(wavelengths)

    # use first order multipole moments (propto local field)
    step_max = torch.max(struct.step)
    enclosing_radius = get_enclosing_sphere_radius(_struct.positions)
    illumination_radius = enclosing_radius + distance_dipoles
    probe_radius = enclosing_radius + distance_probe * step_max
    r0 = struct.get_center_of_mass()

    # setup perpendicular plane waves illuminations
    if n_dipoles is None:
        angles = torch.linspace(-torch.pi, torch.pi * 0.9, 14)
        pw_conf_list = [[1.0, 0.0, angle, "xz"] for angle in angles]
        pw_conf_list += [[0.0, 1.0, angle, "xz"] for angle in angles]
        e_inc_extract = [
            PlaneWave(e0s=a, e0p=b, inc_angle=c, inc_plane=d, device=device)
            for [a, b, c, d] in pw_conf_list
        ]
    # optional: multiple dipole illuminations
    else:
        if n_dipoles <= 0 or type(n_dipoles) != int:
            raise ValueError(
                "dipole illumination mode: `n_dipoles` needs to be a positive integer."
            )

        # setup dipoles of random position and random orientation
        rnd_pos_dp = sample_random_circular(n_dipoles, device=device) * illumination_radius
        e_inc_extract = [
            ElectricLineDipole(
                r_source=r_dp,
                p_source=torch.rand(3, device=device) * illumination_radius,
                device=device,
            )
            for r_dp in rnd_pos_dp
        ]

    # setup field probe positions on enclosing sphere
    r_probe = sample_random_circular(n_probe, device=device) * probe_radius

    # replace illumination
    _sim = Simulation(
        structures=[_struct],
        environment=environment,
        illumination_fields=e_inc_extract,
        wavelengths=wavelengths,
        device=device,
    )

    if verbose:
        t0 = time.time()
        _pos_p, _pos_m = _sim._get_polarizable_positions_separate_p_m()
        n_dp = len(_pos_p) + len(_pos_m)
        n_wl = len(wavelengths)
        print("Running simulation ({} dipoles, {} wls)... ".format(n_dp, n_wl), end="")
    _sim.run(verbose=False, progress_bar=progress_bar)

    if verbose and not progress_bar:
        print("Done in {:.2f}s.".format(time.time() - t0))

    # solve the optimization problem
    alpha_6x6 = torch.zeros(
        (len(wavelengths), 6, 6), dtype=DTYPE_COMPLEX, device=device
    )
    if verbose:
        t0 = time.time()
        print("Running p/m optimization... ", end="")

    for i_wl in tqdm(range(len(wavelengths)), progress_bar, title=""):
        wl = wavelengths[i_wl]

        # calculate fields on enclosing sphere for all illuminations
        # shape: (n_illumination, n_probe*6)
        nf_probe = nf(_sim, wl, r_probe=r_probe, progress_bar=False)
        e_eval = nf_probe["sca"].get_efield()
        h_eval = nf_probe["sca"].get_hfield()
        f_eval = torch.cat([e_eval, h_eval], dim=2)
        f_eval = f_eval.reshape(len(e_inc_extract), -1)

        # illuminating fields at expansion location
        # shape: (n_illumination, 6)
        e0_eval = torch.zeros(
            (len(e_inc_extract), 3), dtype=DTYPE_COMPLEX, device=device
        )
        h0_eval = torch.zeros(
            (len(e_inc_extract), 3), dtype=DTYPE_COMPLEX, device=device
        )
        for i_field, e_inc in enumerate(e_inc_extract):
            inc_f = e_inc.get_field(r0.unsqueeze(0), wl, environment)
            e0_eval[i_field] = inc_f.get_efield()
            h0_eval[i_field] = inc_f.get_hfield()

        f0_eval = torch.cat([e0_eval, h0_eval], dim=1)

        # --- full 6x6 polarizability
        if not only_pE_mH:
            # calculate Green's tensors between r0 and r_probe
            # shape: (n_probe*6, 6)
            G_6x6 = environment.get_G_6x6(r_probe=r_probe, r_source=r0, wavelength=wl)
            G_all = G_6x6.reshape(-1, 6)

            # inv. problem #1: probe fields + Green's tensors --> dipole moments
            Gi = torch.linalg.pinv(G_all)
            pm_eff = torch.matmul(Gi.unsqueeze(0), f_eval.unsqueeze(-1))[..., 0]

            # inv. problem #2: dipole moments + illumination --> effective pola
            pinv_f0 = torch.linalg.pinv(f0_eval)
            alpha_6x6_inv = torch.matmul(pinv_f0, pm_eff)

            # optimum alphas to obtain dipole moments for each illumination
            alpha_6x6[i_wl] = alpha_6x6_inv

            # test residuals
            _test_residual_effective_polarizability(
                _struct,
                wl,
                environment,
                alpha_6x6_inv,
                f0_eval,
                pm_eff,
                text_which_dp="6x6 dipole",
            )

        # --- only pE and mH
        if only_pE_mH:
            # calculate Green's tensors between r0 and r_probe
            # shape: (n_probe*6, 6)
            G_Ep = environment.get_G_Ep(r_probe=r_probe, r_source=r0, wavelength=wl)
            G_Hm = environment.get_G_Hm(r_probe=r_probe, r_source=r0, wavelength=wl)
            G_Ep_all = G_Ep.reshape(-1, 3)
            G_Hm_all = G_Hm.reshape(-1, 3)

            e_eval = e_eval.reshape(len(e_inc_extract), -1)
            h_eval = h_eval.reshape(len(e_inc_extract), -1)

            # inv. problem #1: probe fields + Green's tensors --> dipole moments
            G_Ep_i = torch.linalg.pinv(G_Ep_all)
            G_Hm_i = torch.linalg.pinv(G_Hm_all)
            p_eff = torch.matmul(G_Ep_i.unsqueeze(0), e_eval.unsqueeze(-1))[..., 0]
            m_eff = torch.matmul(G_Hm_i.unsqueeze(0), h_eval.unsqueeze(-1))[..., 0]

            # pseudo-inverse of all illuminations
            pinv_e0 = torch.linalg.pinv(e0_eval)
            pinv_h0 = torch.linalg.pinv(h0_eval)

            # optimum alphas to obtain dipole moments for each illumination
            alpha_pinv = torch.matmul(pinv_e0, p_eff)
            alpha_minv = torch.matmul(pinv_h0, m_eff)

            alpha_6x6[i_wl, :3, :3] = alpha_pinv
            alpha_6x6[i_wl, 3:, 3:] = alpha_minv

            # test residuals
            _test_residual_effective_polarizability(
                _struct,
                wl,
                environment,
                alpha_pinv,
                e0_eval,
                p_eff,
                text_which_dp="electric dipole",
            )
            _test_residual_effective_polarizability(
                _struct,
                wl,
                environment,
                alpha_minv,
                h0_eval,
                m_eff,
                text_which_dp="magnetic dipole",
            )

    if verbose:
        print("Done in {:.2f}s.".format(time.time() - t0))

    dict_pola_pseudo = dict(
        r0=r0,
        r0_MD=r0,
        r0_ED=r0,
        full_geometry=_struct.positions,
        alpha_6x6=alpha_6x6,
        # alpha_pE=alpha_6x6[:, :3, :3],
        # alpha_mH=alpha_6x6[:, 3:, 3:],
        wavelengths=wavelengths,
        enclosing_radius=enclosing_radius,
        k0_spectrum=2 * torch.pi / wavelengths,
    )
    # if not only_pE_mH:
    #     dict_pola_pseudo["alpha_mE"] = alpha_6x6[:, 3:, :3]
    #     dict_pola_pseudo["alpha_pH"] = alpha_6x6[:, :3, 3:]

    return dict_pola_pseudo



# -*- coding: utf-8 -*-
"""
core field container class
"""
# %%
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import tqdm


# Container for fields
class Field:
    """container for electro-magnetic field data

    contains:
     - evaluation points
     - if applicable: line elements or surface elements

    routines:
     - field intensity
     - intensity integration
     - chirality
     - Poynting
     - plotting

    """

    def __init__(
        self,
        positions: torch.Tensor,
        efield: torch.Tensor,
        hfield: torch.Tensor = None,
        ds: torch.Tensor = None,
        device=None,
    ):
        """container for electro-magnetic field data

        Args:
            positions (torch.Tensor): positions at which field data exists. Either a tensor of shape (N,3), or a dict containing "r_probe" and "ds" keys.
            efield (torch.Tensor): complex electric field at positions. Shape must match the positions.
            hfield (torch.Tensor, optional): complex magnetic field at positions. Shape must match the positions. If not given, set to zero (e.g. for far-fields). Defaults to None.
            ds (torch.Tensor, optional): optional surface (or line) elements corresponding to each position. If given, used in field integration. Defaults to None.
            device (str, optional): If None, use device of `efield` tensor. Defaults to None.

        Raises:
            ValueError: insufficient inputs provided
        """
        # exception handling
        if efield is None and hfield is None:
            raise ValueError("At least one of efield or hfield needs to be given.")

        # set device
        if device is None:
            self.device = efield.device
        else:
            self.device = device

        # main data
        self.efield = efield
        self.hfield = hfield
        if type(positions) == dict:
            self.ds = positions["ds"] if (ds is None and "ds" in positions) else ds
            self.positions = positions["r_probe"]
        else:
            self.ds = ds
            self.positions = positions

        # set not given data to zero / unit
        if self.ds is None:  # surface elements (for integration)
            self.ds = torch.ones(
                len(self.positions), device=self.device, dtype=DTYPE_FLOAT
            )
        if self.efield is None:
            self.efield = torch.zeros_like(self.hfield)
        if self.hfield is None:
            self.hfield = torch.zeros_like(self.efield)

        # move tensors to same device as fields
        self.set_device(self.device)

        assert self.efield.shape[1:] == self.positions.shape
        assert self.hfield.shape[1:] == self.positions.shape
        assert len(self.ds) == len(self.positions)

    def set_device(self, device):
        self.positions = self.positions.to(device=device)
        self.efield = self.efield.to(device=device)
        self.ds = self.ds.to(device=device)
        if self.hfield is not None:
            self.hfield = self.hfield.to(device=device)

    def set_efield(self, efield: torch.Tensor):
        """replace the electric field by `efield`"""
        assert efield.shape[1:] == self.positions.shape
        self.efield = torch.as_tensor(efield, device=self.device)

    def set_hfield(self, hfield: torch.Tensor):
        """replace the magnetic field by `hfield`"""
        assert hfield.shape[1:] == self.positions.shape
        self.hfield = torch.as_tensor(hfield, device=self.device)

    def set_ds(self, ds: torch.Tensor):
        """replace the surface elements by `ds`"""
        assert len(ds) == len(self.positions)
        self.ds = torch.as_tensor(ds, device=self.device)

    # --- postprocessing methods
    def get_positions(self, **kwargs):
        """return all positions"""
        
        return self.positions
    
    def get_efield(self, **kwargs):
        """return the electric field at all positions"""
        if torch.sum(torch.abs(self.efield)) == 0:
            warnings.warn("No e-fields available.")

        return self.efield

    def get_hfield(self, **kwargs):
        """return the magnetic field at all positions"""
        if torch.sum(torch.abs(self.hfield)) == 0:
            warnings.warn("No h-fields available.")

        return self.hfield

    def get_efield_intensity(self, **kwargs):
        """return the electric field intensity at all positions"""
        if torch.sum(torch.abs(self.efield)) == 0:
            warnings.warn("No e-fields available.")

        I_e = torch.sum(torch.abs(self.efield) ** 2, dim=-1)
        return I_e

    def get_hfield_intensity(self, **kwargs):
        """return the magnetic field intensity at all positions"""
        if torch.sum(torch.abs(self.hfield)) == 0:
            warnings.warn("No h-fields available.")

        I_h = torch.sum(torch.abs(self.hfield) ** 2, dim=-1)
        return I_h

    def get_poynting(self, **kwargs):
        """get the complex Poynting vector at all positions

        S = conj(E) x H
        (conj(): complex conjugate)
        """
        if torch.sum(torch.abs(self.efield)) == 0:
            warnings.warn("No e-fields available.")
        if torch.sum(torch.abs(self.hfield)) == 0:
            warnings.warn("No h-fields available.")

        S = torch.cross(torch.conj(self.efield), self.hfield, dim=-1)
        return S

    def get_energy_flux(self, **kwargs):
        """time average Poynting vector at all positions

        <S>=Re(S)
        """
        return self.get_poynting().real

    def get_chirality(self, **kwargs):
        """return the near-field chirality at all positions

        calculated as in Meinzer et al, PRB 88, 041407, 2013:
        C ~ Im(conj(E) * B)
        """
        if torch.sum(torch.abs(self.efield)) == 0:
            warnings.warn("No e-fields available.")
        if torch.sum(torch.abs(self.hfield)) == 0:
            warnings.warn("No h-fields available.")

        C = (
            -1
            * torch.sum(
                torch.multiply(torch.conj(self.efield), self.hfield), dim=-1
            ).imag
        )
        return C

    # --- integration methods
    def get_integrated_efield_intensity(self, **kwargs):
        """get the integrated electric field intensity

        return sum(abs(E)^2 x dS) as a single float value
        """
        if torch.mean(torch.abs(self.efield)) == 1:
            warnings.warn("Surface elements are set to 1.")
        return torch.sum(self.get_efield_intensity() * self.ds, dim=-1)

    def get_integrated_hfield_intensity(self, **kwargs):
        """get the integrated magnetic field intensity

        return sum(abs(H)^2 x dS) as a single float value
        """
        if torch.mean(torch.abs(self.efield)) == 1:
            warnings.warn("Surface elements are set to 1.")
        return torch.sum(self.get_hfield_intensity() * self.ds, dim=-1)

    # --- plotting methods
    # - plotting for e-field
    def plot_efield_amplitude(
        self, illumination_index=0, field_component="x", complex_part="re", **kwargs
    ):
        """plot E-field amplitude using :func:`~torchgdm.visu.visu2d.field_amplitude`"""
        from torchgdm.visu.visu2d import field_amplitude as visu_field_amplitude_2d

        return visu_field_amplitude_2d(
            self,
            illumination_index=illumination_index,
            whichfield="e",
            field_component=field_component,
            complex_part=complex_part,
            **kwargs
        )

    def plot_efield_vectors(self, illumination_index=0, **kwargs):
        """plot E-field vectors using :func:`~torchgdm.visu.visu2d.vectorfield`"""
        from torchgdm.visu.visu2d import vectorfield as visu_vectorfield_2d

        return visu_vectorfield_2d(
            self, illumination_index=illumination_index, whichfield="e", **kwargs
        )

    def plot_efield_intensity(self, illumination_index=0, **kwargs):
        """plot E-field intensity map using :func:`~torchgdm.visu.visu2d.field_intensity`"""
        from torchgdm.visu.visu2d import field_intensity as visu_field_intensity_2d

        return visu_field_intensity_2d(
            self, illumination_index=illumination_index, whichfield="e", **kwargs
        )

    def plot_efield_vectors3d(self, illumination_index=0, **kwargs):
        """plot E-field 3D vectors using :func:`~torchgdm.visu.visu3d.vectorfield`"""
        from torchgdm.visu.visu3d import vectorfield as visu_vectorfield_3d

        return visu_vectorfield_3d(
            self, illumination_index=illumination_index, whichfield="e", **kwargs
        )

    # - plotting for h-field
    def plot_hfield_amplitude(
        self, illumination_index=0, field_component="x", complex_part="re", **kwargs
    ):
        """plot H-field amplitude using :func:`~torchgdm.visu.visu2d.field_amplitude`"""
        from torchgdm.visu.visu2d import field_amplitude as visu_field_amplitude_2d

        return visu_field_amplitude_2d(
            self,
            illumination_index=illumination_index,
            whichfield="h",
            field_component=field_component,
            complex_part=complex_part,
            **kwargs
        )

    def plot_hfield_vectors(self, illumination_index=0, **kwargs):
        """plot H-field vectors using :func:`~torchgdm.visu.visu2d.vectorfield`"""
        from torchgdm.visu.visu2d import vectorfield as visu_vectorfield_2d

        return visu_vectorfield_2d(
            self, illumination_index=illumination_index, whichfield="h", **kwargs
        )

    def plot_hfield_intensity(self, illumination_index=0, **kwargs):
        """plot H-field intensity map using :func:`~torchgdm.visu.visu2d.field_intensity`"""
        from torchgdm.visu.visu2d import field_intensity as visu_field_intensity_2d

        return visu_field_intensity_2d(
            self, illumination_index=illumination_index, whichfield="h", **kwargs
        )

    def plot_hfield_vectors3d(self, illumination_index=0, **kwargs):
        """plot H-field 3D vectors using :func:`~torchgdm.visu.visu3d.vectorfield`"""
        from torchgdm.visu.visu3d import vectorfield as visu_vectorfield_3d

        return visu_vectorfield_3d(
            self, illumination_index=illumination_index, whichfield="h", **kwargs
        )

    # - plot chirality
    def plot_chirality(self, illumination_index=0, **kwargs):
        """plot nearfield chirality :func:`~torchgdm.visu.visu2d.`"""
        from torchgdm.visu.visu2d.scalar2d import _scalarfield

        return _scalarfield(
            self.get_chirality()[illumination_index], self.positions, **kwargs
        )

    # - plotting for time averaged Poynting
    def plot_energy_flux_vectors(self, illumination_index=0, **kwargs):
        """plot energy flux vector field using :func:`torchgdm.visu.visu2d.vec2d._vectorfield`"""
        from torchgdm.visu.visu2d.vec2d import _vectorfield

        f = self.get_energy_flux()[illumination_index]

        return _vectorfield(f, self.positions, **kwargs)

    def plot_energy_flux_streamlines(self, illumination_index=0, **kwargs):
        """plot energy flux streamlines using :func:`~torchgdm.visu.visu2d.vec2d.streamlines_energy_flux`"""
        from torchgdm.visu.visu2d.vec2d import streamlines_energy_flux

        return streamlines_energy_flux(
            self, illumination_index=illumination_index, **kwargs
        )



# -*- coding: utf-8 -*-
"""
provider for dielectric functions of selected isotropic materials.
"""
# %%
import warnings
import importlib.resources as pkg_resources
import pathlib

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm import materials  # caution, dangerous wrt circular imports
from torchgdm.materials.base_classes import MaterialBase
from torchgdm.tools.misc import to_np
from torchgdm.tools.interp import interp1d

# --- get all available materials
DATA_FOLDER = "data/"
data_files = pkg_resources.files(materials).joinpath(DATA_FOLDER).iterdir()

REFINDEX_DATA = {}
for f in data_files:
    f_n = pathlib.Path(f).name
    mat_name = f_n.split("_")[0]
    REFINDEX_DATA[mat_name.lower()] = [f, mat_name]


def list_available_materials(verbose=False):
    if verbose:
        for f in REFINDEX_DATA:
            print("{}: ".format(f, pathlib.Path(REFINDEX_DATA[f]).name))
    return [f for f in REFINDEX_DATA]


# --- internal helper
def _load_tabulated(dat_str):
    rows = dat_str["data"].split("\n")
    splitrows = [c.split() for c in rows]
    wl = []
    eps = []
    for s in splitrows:
        if len(s) > 0:
            wl.append(1000.0 * float(s[0]))  # microns --> nm
            _n = float(s[1])
            if len(s) > 2:
                _k = float(s[2])
            else:
                _k = 0.0
            eps.append((_n + 1j * _k) ** 2)
    return wl, eps


def _load_formula(dat_str):
    model_type = int((dat_str["type"].split())[1])
    coeff = [float(s) for s in dat_str["coefficients"].split()]
    for k in ["range", "wavelength_range"]:
        if k in dat_str:
            break
    # validity range (convert to nm)
    wl_range = [1e3 * float(dat_str[k].split()[0]), 1e3 * float(dat_str[k].split()[1])]

    return model_type, wl_range, coeff


# --- main interface classes
class MatDatabase(MaterialBase):
    """dispersion from a database entry

    Use permittivity data from included database (data from https://refractiveindex.info/),
    or by loading a yaml file downloaded from https://refractiveindex.info/. Currently
    supported ref.index formats are tabulated n(k) data or Sellmeier model.

    Tabulated materials natively available in torchgdm can be
    printed via :func:`torchgdm.materials.list_available_materials()`

    Requires `pyyaml` (pip3 install pyyaml)

    Parameters
    ----------
    name : str
        name of database entry

    yaml_file : str, default: None
        optional filename of yaml refractiveindex data to load. In case a
        filename is provided, `name` will only be used as __name__ attribute
        for the class instance.

    """

    def __init__(
        self,
        name="",
        yaml_file=None,
        device: torch.device = None,
        init_lookup_wavelengths=None,
    ):
        """dispersion from a database entry

        supports data following the yaml format of refractiveindex.info.
        Currently tabulated permittivity and Sellmeier models are supported.

        Args:
            name (str, optional): Name of database entry. Defaults to "".
            yaml_file (_type_, optional): path to optional yaml file with material data to load. If given, file is loaded and no data-base entry will be used, even if the name matches. Defaults to None.
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).
            init_lookup_wavelengths (torch.Tensor, optional): optional list of wavelengths to generate an initial lookup table. Defaults to None.

        Raises:
            ValueError: Unknown material or unknown dispersion model type
        """
        import yaml

        super().__init__(device=device)

        if (name == "") and (yaml_file is None):
            print("No material specified. Available materials in database: ")
            for k in REFINDEX_DATA:
                print("     - '{}'".format(k))
            del self
            return
        if (yaml_file is None) and (name.lower() not in REFINDEX_DATA):
            raise ValueError(
                "'{}': Unknown material. Available materials in database: {}".format(
                    name, REFINDEX_DATA.keys()
                )
            )

        # load database entry from yaml
        if yaml_file is None:
            yaml_file = REFINDEX_DATA[name.lower()][0]
            self.__name__ = REFINDEX_DATA[name.lower()][1]
        else:
            if name:
                self.__name__ = name
            else:
                self.__name__ = pathlib.Path(yaml_file).stem

        with open(yaml_file, "r", encoding="utf8") as f:
            self.dset = yaml.load(f, Loader=yaml.BaseLoader)

        if len(self.dset["DATA"]) > 1:
            warnings.warn(
                "Several model entries in data-set for '{}' ({}). Using first entry.".format(
                    name, yaml_file
                )
            )
        dat = self.dset["DATA"][0]
        self.type = dat["type"]
        self.wl_dat = torch.Tensor([])
        self.eps_dat = torch.Tensor([])
        self.lookup_eps = {}

        # load refractive index model.
        # currently supported: tabulated data and Sellmeier model.
        # - tabulated data
        if self.type.split()[0] == "tabulated":
            wl_dat, eps_dat = _load_tabulated(dat)
            self.wl_dat = torch.as_tensor(wl_dat, dtype=DTYPE_FLOAT, device=self.device)
            self.eps_dat = torch.as_tensor(
                eps_dat, dtype=DTYPE_COMPLEX, device=self.device
            )
            self.model_type = "data"
            self.coeff = []
            self.wl_range = [torch.min(self.wl_dat), torch.max(self.wl_dat)]

        # - Sellmeier
        elif self.type.split()[0] == "formula":
            self.model_type, self.wl_range, self.coeff = _load_formula(dat)
            if self.model_type == 1:
                self.model_type = "sellmeier"
        else:
            raise ValueError(
                "refractiveindex.info data type '{}' not implemented yet.".format(
                    self.type
                )
            )

        # optionally initialize wavelength lookup
        if init_lookup_wavelengths is not None:
            for wl in init_lookup_wavelengths:
                _eps = self._get_eps_single_wl(wl)

    def __repr__(self, verbose: bool = False):
        """description about material"""
        out_str = ' ----- Material "{}" ({}) -----'.format(
            self.__name__, self.model_type
        )
        if self.model_type == "data":
            out_str += "\n tabulated wavelength range: {:.1f}nm - {:.1f}nm".format(
                *self.wl_range
            )
        elif self.model_type == "sellmeier":
            out_str += "\n Sellmeier model validity range: {:.1f}nm - {:.1f}nm".format(
                *self.wl_range
            )
        return out_str

    def set_device(self, device):
        super().set_device(device)
        self.wl_dat = self.wl_dat.to(device)
        self.eps_dat = self.eps_dat.to(device)

        # transfer the lookup table
        for _w in self.lookup_eps:
            self.lookup_eps[_w] = self.lookup_eps[_w].to(device)

    def _eval(self, wavelength):
        """evaluate refractiveindex.info model"""

        # - tabulated, using bilinear interpolation
        if self.model_type == "data":
            # torch implementation of 1d interpolation:
            eps = interp1d(wavelength, self.wl_dat, self.eps_dat)

        # - Sellmeier
        elif self.model_type == "sellmeier":
            eps = 1 + self.coeff[0]

            def g(c1, c2, w):
                return c1 * (w**2) / (w**2 - c2**2)

            for i in range(1, len(self.coeff), 2):
                # wavelength factor 1/1000: nm --> microns
                wl_mu = wavelength / 1000.0
                eps += g(self.coeff[i], self.coeff[i + 1], wl_mu)

        else:
            raise ValueError(
                "Only formula '1' (Sellmeier) or 'data' models supported so far."
            )

        return eps

    def _get_eps_single_wl(self, wavelength):
        # memoize evaluations
        wl_key = float(wavelength)

        if wl_key in self.lookup_eps:
            eps = self.lookup_eps[wl_key]
        else:
            _eps = self._eval(wavelength)
            eps = torch.eye(
                3, dtype=DTYPE_COMPLEX, device=self.device
            ) * torch.as_tensor(_eps, dtype=DTYPE_COMPLEX, device=self.device)
            self.lookup_eps[wl_key] = eps

        return eps

    def get_epsilon(self, wavelength):
        """get permittivity at `wavelength`

        Args:
            wavelength (float): in nm

        Returns:
            torch.Tensor: (3,3) complex permittivity tensor at `wavelength`
        """
        wavelength = torch.as_tensor(wavelength, dtype=DTYPE_FLOAT, device=self.device)

        # multiple wavelengths
        if len(torch.as_tensor(wavelength).shape) == 1:
            eps = torch.stack([self._get_eps_single_wl(wl) for wl in wavelength], dim=0)
        else:
            eps = self._get_eps_single_wl(wavelength)

        return eps



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestSimCrossSections(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing multipole decomposition...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        eps_env = 1.0
        mat_env = tg.materials.MatConstant(eps_env)
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

        # - illumination field(s)
        self.wl = torch.as_tensor(720.0)
        self.wavelengths = torch.as_tensor([self.wl])
        self.e_inc_list = [
            tg.env.freespace_3d.PlaneWave(e0p=1.0, e0s=0.0, inc_angle=0),
        ]

        # - structure
        step = 33.0  # nm
        mat_struct = tg.materials.MatConstant(eps=12.0)
        self.struct1 = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(3, 3, 2),
            step,
            mat_struct,
            radiative_correction=False,
        )

        self.sim = tg.simulation.Simulation(
            structures=[self.struct1],
            environment=self.env,
            illumination_fields=self.e_inc_list,
            wavelengths=self.wavelengths,
        )

    def test_effective_polarizability(self):
        for device in self.devices:
            self.sim.set_device(device)
            # random seed for reproducible extraction (based on rnd dipole sources)
            torch.manual_seed(1)

            # extract effective ed/md polarizabilities
            a_tgdm = tg.struct.volume.pola.extract_eff_pola_via_mp_decomposition(
                self.struct1,
                self.env,
                self.wavelengths,
                verbose=False,
                progress_bar=False,
            )

            # create equivalent effective pola simulation
            struct_eff = tg.struct.StructEffPola3D(
                a_tgdm["r0"].unsqueeze(0), a_tgdm
            )
            sim_aeff = tg.Simulation(
                structures=[struct_eff],
                environment=self.env,
                illumination_fields=self.e_inc_list,
                wavelengths=self.wavelengths,
                device=device,
            )

            # run both simulations
            self.sim.run(verbose=False, progress_bar=False)
            sim_aeff.run(verbose=False, progress_bar=False)

            # --- compare scat/extinct of full and effective pola. simulations
            cs_full = tg.postproc.crosssect.total(self.sim, self.wl)
            cs_aeff = tg.postproc.crosssect.total(sim_aeff, self.wl)
            # print(cs_full["scs"], cs_aeff["scs"])
            # print(cs_full["ecs"], cs_aeff["ecs"])

            torch.testing.assert_close(
                cs_full["scs"], cs_aeff["scs"], atol=1e4, rtol=2e-2
            )
            torch.testing.assert_close(
                cs_full["ecs"], cs_aeff["ecs"], atol=1e4, rtol=2e-2
            )

            if self.verbose:
                print(
                    "  - {}: effective polarizability extraction test passed.".format(
                        device
                    )
                )

            # --- compare total scat/extinct with mutlipole sum
            mp_full_s = tg.postproc.multipole.scs(self.sim, self.wl)
            scs_ed = mp_full_s["scs_ed"]
            scs_md = mp_full_s["scs_md"]
            scs_eq = mp_full_s["scs_eq"]
            scs_mq = mp_full_s["scs_mq"]
            scs_mp_sum = scs_ed + scs_md + scs_eq + scs_mq

            mp_full_e = tg.postproc.multipole.ecs(self.sim, self.wl)
            ecs_ed = mp_full_e["ecs_ed"]
            ecs_md = mp_full_e["ecs_md"]
            ecs_eq = mp_full_e["ecs_eq"]
            ecs_mq = mp_full_e["ecs_mq"]
            ecs_mp_sum = ecs_ed + ecs_md + ecs_eq + ecs_mq
            # print(cs_full["scs"], scs_mp_sum)
            # print(cs_full["ecs"], ecs_mp_sum)

            torch.testing.assert_close(
                cs_full["scs"], scs_mp_sum, atol=5e3, rtol=1e-2
            )
            torch.testing.assert_close(
                cs_full["ecs"], ecs_mp_sum, atol=5e3, rtol=1e-2
            )

            if self.verbose:
                print("  - {}: multipole sum test passed.".format(device))


# %%
if __name__ == "__main__":
    print("testing exact multipole decomposition...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestSim3DSplitCombine(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("test splitting and combining simulations...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        eps_env = 1.0
        mat_env = tg.materials.MatConstant(eps_env)
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

        # - illumination field(s)
        self.wl = 733.0
        self.e_inc_list = [
            tg.env.freespace_3d.PlaneWave(e0p=0.8, e0s=1.0, inc_angle=torch.pi / 3),
        ]

        # - structure
        step = 25.0
        mat = tg.materials.MatDatabase("gaas")
        self.s1 = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(l=2, w=1, h=1),
            step,
            mat,
        )
        step = 15.0
        mat = tg.materials.MatDatabase("Si")
        self.s2 = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(l=1, w=2, h=1),
            step,
            mat,
        )

        self.sim_conf = dict(
            environment=self.env,
            illumination_fields=self.e_inc_list,
            wavelengths=torch.tensor([self.wl]),
        )

    def test_combine_sim_struct(self):
        for device in self.devices:
            s1 = self.s1 - [100, 200, 50]
            s2 = self.s2 + [100, -50, 150]
            sim1 = tg.simulation.Simulation(
                structures=[s1], device=device, **self.sim_conf
            )
            sim2 = tg.simulation.Simulation(
                structures=[s2], device=device, **self.sim_conf
            )
            # combine the two structures / simulations in different ways
            sim_comb_list = []
            sim_comb_list.append(
                tg.simulation.Simulation(
                    structures=[s1, s2], device=device, **self.sim_conf
                )
            )
            sim_comb_list.append(
                tg.simulation.Simulation(
                    structures=[s1.combine(s2)], device=device, **self.sim_conf
                )
            )
            sim_comb_list.append(
                tg.simulation.Simulation(
                    structures=[s1 + s2], device=device, **self.sim_conf
                )
            )
            sim_comb_list.append(sim1 + sim2)
            sim_comb_list.append(sim1.combine(sim2))

            # test if combined simulations are same
            for _s1 in sim_comb_list:
                for _s2 in sim_comb_list:
                    torch.testing.assert_close(
                        _s1.get_all_positions(), _s2.get_all_positions()
                    )
                    torch.testing.assert_close(
                        _s1.get_source_validity_radius(), _s2.get_source_validity_radius()
                    )
                    torch.testing.assert_close(
                        _s1._get_all_polarizabilitites_6x6(self.wl),
                        _s2._get_all_polarizabilitites_6x6(self.wl),
                    )

    def test_run_seperate_and_combined(self):
        for device in self.devices:

            self.s1.set_center_of_mass([-100, 0, 0])
            self.s2.set_center_of_mass([100, 50, 0])

            sim1 = tg.simulation.Simulation(
                structures=[self.s1, self.s2],
                device=device,
                **self.sim_conf,
            )
            sim1.run(calc_missing=True, verbose=False, progress_bar=False)

            sim2 = tg.simulation.Simulation(
                structures=[self.s1 + self.s2],
                device=device,
                **self.sim_conf,
            )
            sim2.run(calc_missing=True, verbose=False, progress_bar=False)

            torch.testing.assert_close(
                sim1.fields_inside[self.wl].efield,
                sim2.fields_inside[self.wl].efield,
            )

            torch.testing.assert_close(
                sim1.fields_inside[self.wl].hfield,
                sim2.fields_inside[self.wl].hfield,
            )

            if self.verbose:
                print("  - {}: list vs combined struct 3D sim test passed.".format(device))

    def test_calculate_E_H(self):
        for device in self.devices:

            self.s1.set_center_of_mass([-100, 0, 0])
            self.s2.set_center_of_mass([100, 50, 0])

            sim = tg.simulation.Simulation(
                structures=[self.s1, self.s2],
                device=device,
                **self.sim_conf,
            )

            sim.run(calc_missing=True, verbose=False, progress_bar=False)

            # test split/combine
            s_split0 = sim.split(0)
            s_split1 = sim.split(1)
            s_comb = s_split0.combine(s_split1)

            torch.testing.assert_close(
                sim.fields_inside[self.wl].efield,
                s_comb.fields_inside[self.wl].efield,
            )

            torch.testing.assert_close(
                sim.fields_inside[self.wl].hfield,
                s_comb.fields_inside[self.wl].hfield,
            )

            # test delete
            s0_delete = sim.delete_struct(1)
            s1_delete = sim.delete_struct(0)

            torch.testing.assert_close(
                s_split0.fields_inside[self.wl].efield,
                s0_delete.fields_inside[self.wl].efield,
            )

            torch.testing.assert_close(
                s_split1.fields_inside[self.wl].hfield,
                s1_delete.fields_inside[self.wl].hfield,
            )

            if self.verbose:
                print("  - {}: split combine 3D sim test passed.".format(device))


class TestSim2DSplitCombine(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("test splitting and combining simulations...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        eps_env = 1.0
        mat_env = tg.materials.MatConstant(eps_env)
        self.env = tg.env.freespace_2d.EnvHomogeneous2D(env_material=mat_env)

        # - illumination field(s)
        self.wl = 670.0
        self.e_inc_list = [
            tg.env.freespace_2d.PlaneWave(
                e0p=0.9j, e0s=0.4, inc_angle=-torch.pi / 1.55
            ),
        ]

        # - structure
        step = 25.0
        mat = tg.materials.MatDatabase("gap")
        self.s1 = tg.struct.StructDiscretizedSquare2D(
            tg.struct.surface_2d.rectangle(l=3, h=2), step, mat
        )
        step = 20.0
        mat = tg.materials.MatDatabase("ge")
        self.s2 = tg.struct.StructDiscretizedSquare2D(
            tg.struct.surface_2d.rectangle(l=1, h=3), step, mat
        )

        self.sim_conf = dict(
            environment=self.env,
            illumination_fields=self.e_inc_list,
            wavelengths=torch.tensor([self.wl]),
        )

    def test_combine_sim_struct(self):
        for device in self.devices:
            s1 = self.s1 - [100, 0, 50]
            s2 = self.s2 + [100, 0, 150]
            sim1 = tg.simulation.Simulation(
                structures=[s1], device=device, **self.sim_conf
            )
            sim2 = tg.simulation.Simulation(
                structures=[s2], device=device, **self.sim_conf
            )
            # combine the two structures / simulations in different ways
            sim_comb_list = []
            sim_comb_list.append(
                tg.simulation.Simulation(
                    structures=[s1, s2], device=device, **self.sim_conf
                )
            )
            sim_comb_list.append(
                tg.simulation.Simulation(
                    structures=[s1.combine(s2)], device=device, **self.sim_conf
                )
            )
            sim_comb_list.append(
                tg.simulation.Simulation(
                    structures=[s1 + s2], device=device, **self.sim_conf
                )
            )
            sim_comb_list.append(sim1 + sim2)
            sim_comb_list.append(sim1.combine(sim2))

            # test if combined simulations are same
            for _s1 in sim_comb_list:
                for _s2 in sim_comb_list:
                    torch.testing.assert_close(
                        _s1.get_all_positions(), _s2.get_all_positions()
                    )
                    torch.testing.assert_close(
                        _s1.get_source_validity_radius(), _s2.get_source_validity_radius()
                    )
                    torch.testing.assert_close(
                        _s1._get_all_polarizabilitites_6x6(self.wl),
                        _s2._get_all_polarizabilitites_6x6(self.wl),
                    )

    def test_run_seperate_and_combined(self):
        for device in self.devices:

            self.s1.set_center_of_mass([-100, 0, 0])
            self.s2.set_center_of_mass([100, 0, 20])

            sim1 = tg.simulation.Simulation(
                structures=[self.s1, self.s2],
                device=device,
                **self.sim_conf,
            )
            sim1.run(calc_missing=True, verbose=False, progress_bar=False)

            sim2 = tg.simulation.Simulation(
                structures=[self.s1 + self.s2],
                device=device,
                **self.sim_conf,
            )
            sim2.run(calc_missing=True, verbose=False, progress_bar=False)

            torch.testing.assert_close(
                sim1.fields_inside[self.wl].efield,
                sim2.fields_inside[self.wl].efield,
            )

            torch.testing.assert_close(
                sim1.fields_inside[self.wl].hfield,
                sim2.fields_inside[self.wl].hfield,
            )

            if self.verbose:
                print("  - {}: list vs combined struct 2D sim test passed.".format(device))

    def test_calculate_E_H(self):
        for device in self.devices:

            self.s1.set_center_of_mass([-100, 0, 0])
            self.s2.set_center_of_mass([100, 0, 50])

            sim = tg.simulation.Simulation(
                structures=[self.s1, self.s2], device=device, **self.sim_conf
            )

            sim.run(calc_missing=True, verbose=False, progress_bar=False)

            # test split/combine
            s_split0 = sim.split(0)
            s_split1 = sim.split(1)
            s_comb = s_split0.combine(s_split1)

            torch.testing.assert_close(
                sim.fields_inside[self.wl].efield,
                s_comb.fields_inside[self.wl].efield,
            )

            torch.testing.assert_close(
                sim.fields_inside[self.wl].hfield,
                s_comb.fields_inside[self.wl].hfield,
            )

            # test delete
            s0_delete = sim.delete_struct(1)
            s1_delete = sim.delete_struct(0)

            torch.testing.assert_close(
                s_split0.fields_inside[self.wl].efield,
                s0_delete.fields_inside[self.wl].efield,
            )

            torch.testing.assert_close(
                s_split1.fields_inside[self.wl].hfield,
                s1_delete.fields_inside[self.wl].hfield,
            )

            if self.verbose:
                print("  - {}: split combine 2D sim test passed.".format(device))


# %%
if __name__ == "__main__":
    print("testing splitting / combining simulations...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""
provider for dielectric functions of selected materials.
"""
# %%
import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.materials.base_classes import MaterialBase


class MatConstant(MaterialBase):
    """constant material index

    Material without dispersion
    """

    def __init__(self, eps=2.0 + 0.0j, device: torch.device = None):
        """constant permittivity material

        Args:
            eps (complex, optional): complex permittivity value. Defaults to (2.0 + 0.0j).
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).
        """
        super().__init__(device=device)

        self.eps_scalar = torch.as_tensor(eps, dtype=DTYPE_COMPLEX, device=self.device)

        _eps_re = torch.round(self.eps_scalar.real, decimals=1)
        _eps_im = torch.round(self.eps_scalar.imag, decimals=3)
        if _eps_im == 0:
            self.__name__ = "eps={}".format(_eps_re)
        else:
            self.__name__ = "eps={}+i{}".format(_eps_re, _eps_im)

    def __repr__(self, verbose: bool = False):
        """description about material"""
        out_str = "constant, isotropic material. permittivity = {}".format(
            self.eps_scalar
        )
        return out_str

    def set_device(self, device):
        super().set_device(device)
        self.eps_scalar = self.eps_scalar.to(device)

    def get_epsilon(self, wavelength):
        """dispersionless, constant permittivity function

        Args:
            wavelength (float): in nm

        Returns:
            torch.Tensor: (3,3) complex permittivity tensor at `wavelength`
        """
        eps = torch.eye(3, dtype=DTYPE_COMPLEX, device=self.device)
        wavelength = torch.as_tensor(wavelength, device=self.device)

        # multiple wavelengths
        if len(wavelength.shape) == 1:
            eps = torch.repeat_interleave(
                torch.eye(3, dtype=DTYPE_COMPLEX, device=self.device).unsqueeze(0),
                wavelength.shape[0],
                dim=0,
            )

        return eps * self.eps_scalar


class MatTiO2(MaterialBase):
    """TiO2 permittivity

    model for visible and NIR range (~500nm - ~1500nm)
    from https://refractiveindex.info/?shelf=main&book=TiO2&page=Devore-o

    Thanks to Dr. Frank Mersch (Kronos International) for help
    """

    __name__ = "TiO2 (Devore)"

    def __init__(self, orientation="avg", device: torch.device = None):
        """TiO2 permittivity

        supports single axis, averaged or tensorial permittivity models.

        Args:
            orientation (str, optional): one of 'avg' (average of n_o and n_e: n = (2*n_o + n_e) / 3), 'n_o' (ordinary axis permittivity),          - 'n_e' (extra-ordinary axis permittivity), or 'tensor'. x,y: ordinary axes; z: extraordinary axis. Defaults to "avg".
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).
        """
        super().__init__(device=device)

        self.orientation = orientation.lower()

    def __repr__(self, verbose: bool = False):
        """description about material"""
        out_str = "TiO2 birefringent permittivity. ordinary: x,y; extraordinary: z."
        return out_str

    def set_device(self, device):
        super().set_device(device)

    def _n_o(self, wavelength):
        n = torch.sqrt(
            torch.as_tensor(
                5.913 + 0.2441 / ((wavelength / 1e3) ** 2 - 0.0803),
                dtype=DTYPE_FLOAT,
                device=self.device,
            )
        )
        return n

    def _n_e(self, wavelength):
        n = torch.sqrt(
            torch.as_tensor(
                7.197 + 0.3322 / ((wavelength / 1e3) ** 2 - 0.0843),
                dtype=DTYPE_FLOAT,
                device=self.device,
            )
        )
        return n

    def _get_eps_single_wl(self, wavelength):
        if self.orientation in ["n_o", "no"]:
            # purely real in available spectral range
            eps = (self._n_o(wavelength) ** 2 + 0j) * torch.eye(
                3, dtype=DTYPE_COMPLEX, device=self.device
            )
        elif self.orientation in ["n_e", "ne"]:
            eps = (self._n_e(wavelength) ** 2 + 0j) * torch.eye(
                3, dtype=DTYPE_COMPLEX, device=self.device
            )
        elif self.orientation in ["n_e", "ne"]:
            eps = (self._n_e(wavelength) ** 2 + 0j) * torch.eye(
                3, dtype=DTYPE_COMPLEX, device=self.device
            )
        elif self.orientation == "avg":
            n_o = self._n_o(wavelength)
            n_e = self._n_e(wavelength)
            n_avg = (2 * n_o + n_e) / 3.0
            eps = (n_avg**2 + 0j) * torch.eye(
                3, dtype=DTYPE_COMPLEX, device=self.device
            )
        else:
            n_o = self._n_o(wavelength)
            n_e = self._n_e(wavelength)
            eps = torch.eye(3, dtype=DTYPE_COMPLEX, device=self.device)
            eps[0, 0] = n_o**2
            eps[1, 1] = n_o**2
            eps[2, 2] = n_e**2

        return eps

    def get_epsilon(self, wavelength):
        """get permittivity at `wavelength`

        Args:
            wavelength (float): in nm

        Returns:
            torch.Tensor: (3,3) complex permittivity tensor at `wavelength`
        """

        # multiple wavelengths
        if len(torch.as_tensor(wavelength).shape) == 1:
            eps = torch.stack([self._get_eps_single_wl(wl) for wl in wavelength], dim=0)
        else:
            eps = self._get_eps_single_wl(wavelength)

        return eps



# -*- coding: utf-8 -*-
"""
definitions of some global constants    
"""
# %%
import itertools
import warnings

import torch

# --- general configuration
# if True: raise error if wavelength not defined.
# if False: Warn and use closest match
ERROR_ON_WAVELENGTH_MISSMATCH = False


# --- default data types
DTYPE_FLOAT = torch.float32
DTYPE_COMPLEX = torch.complex64
# DTYPE_FLOAT = torch.float64
# DTYPE_COMPLEX = torch.complex128


# --- default torch device
DEFAULT_DEVICE = "cpu"


# --- default matplotlib colors. Generate with:
#     import matplotlib.colors as mcolors
#     COLORS_DEFAULT = [mcolors.to_hex(f"C{i}") for i in range(10)]
COLORS_DEFAULT = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
] * 10  # dirty hack: repeat enough times

# --- structure counter
STRUCTURE_IDS = itertools.cycle(range(100000000000))



# encoding=utf-8
"""environment for 2D structures in a homogeneous environment.
"""
from . import inc_fields

from .dyads import EnvHomogeneous2D

from .inc_fields import NullField
from .inc_fields import PlaneWave
from .inc_fields import ElectricLineDipole
from .inc_fields import MagneticLineDipole


# -*- coding: utf-8 -*-
"""
provider for dielectric functions of selected materials.
"""
# %%
import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device


# --- material defining base class
class MaterialBase:
    """base class for material permittivity"""

    __name__ = "material dielectric constant base class"

    def __init__(self, device: torch.device = None):
        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

    def __repr__(self, verbose: bool = False):
        """description about material"""
        out_str = " ------ base material class - doesn't define anything yet -------"
        return out_str

    def set_device(self, device):
        """move all tensors of the class to device"""
        self.device = device

    def get_epsilon(self, wavelength: float):
        """return (3, 3) permittivity tensor at `wavelength`"""
        raise NotImplementedError("Needs to be implemented in child class.")

    def plot_epsilon(
        self, wavelengths=torch.linspace(400, 1400, 100), tensor_comp=[0, 0]
    ):
        """plot the permittivity dispersion

        Args:
            wavelengths (torch.Tensor, optional): wavelengths to evaluate. Defaults to torch.linspace(400, 1400, 100).
            tensor_comp (list, optional): permittivity tensor component indices. Defaults to [0, 0].
        """
        import matplotlib.pyplot as plt
        from torchgdm import to_np
        from torchgdm.visu.visu2d._tools import _get_axis_existing_or_new

        eps = self.get_epsilon(wavelengths)

        # plot
        ax, show = _get_axis_existing_or_new()

        plt.title("epsilon - '{}'".format(self.__name__))
        plt.plot(
            to_np(wavelengths),
            to_np(eps[:, tensor_comp[0], tensor_comp[1]].real),
            label=r"Re($\epsilon$)",
        )
        plt.plot(
            to_np(wavelengths),
            to_np(eps[:, tensor_comp[0], tensor_comp[1]].imag),
            label=r"Im($\epsilon$)",
        )
        plt.legend()
        plt.xlabel("wavelength (nm)")
        plt.ylabel("permittivity")

        if show:
            plt.show()

    def plot_refractive_index(
        self, wavelengths=torch.linspace(400, 1400, 100), tensor_comp=[0, 0]
    ):
        """plot the refractive index dispersion

        Args:
            wavelengths (torch.Tensor, optional): wavelengths to evaluate. Defaults to torch.linspace(400, 1400, 100).
            tensor_comp (list, optional): refractive index tensor component indices. Defaults to [0, 0].
        """
        import matplotlib.pyplot as plt
        from torchgdm import to_np
        from torchgdm.visu.visu2d._tools import _get_axis_existing_or_new

        n_mat = self.get_epsilon(wavelengths) ** 0.5

        # plot
        ax, show = _get_axis_existing_or_new()

        plt.title("ref. index - '{}'".format(self.__name__))
        plt.plot(
            to_np(wavelengths),
            to_np(n_mat[:, tensor_comp[0], tensor_comp[1]].real),
            label="n",
        )
        plt.plot(
            to_np(wavelengths),
            to_np(n_mat[:, tensor_comp[0], tensor_comp[1]].imag),
            label="k",
        )
        plt.legend()
        plt.xlabel("wavelength (nm)")
        plt.ylabel("refractive index")

        if show:
            plt.show()



# encoding=utf-8
#
# Copyright (C) 2023-2024, P. R. Wiecha
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
torchgdm - a full-field electrodynamical solver written in pytorch.
Based on the Green Dyadic Method.
"""



__name__ = "torchgdm"
__version__ = "0.45"
__date__ = "11/18/2024"  # MM/DD/YYY
__license__ = "GPL3"
__status__ = "beta"

__copyright__ = "Copyright 2023-2024, Peter R. Wiecha"
__author__ = "Peter R. Wiecha"
__maintainer__ = "Peter R. Wiecha"
__email__ = "pwiecha@laas.fr"
__credits__ = [
    "Christian Girard",
    "Arnaud Arbouet",
    "Clément Majorel",
    "Sofia Ponomareva",
    "Antoine Azéma",
    "Adelin Patoux",
    "Renaud Marty",
]

# --- populate namespace
from .constants import DEFAULT_DEVICE

device = DEFAULT_DEVICE

# make some functions and classes available at top level
from .simulation import Simulation
from .field import Field
from .tools.misc import to_np, tqdm
from .tools.misc import use_cuda, get_default_device, set_default_device

# make some submodules available at top level
from .visu import visu2d
from .visu import visu3d

# modules
from . import constants
from . import linearsystem
from . import simulation
from . import field

# sub packages
from . import env
from . import materials
from . import struct
from . import postproc
from . import tools



# encoding=utf-8
# %%
import unittest
import warnings

import torch

import torchgdm as tg


class TestDimensionMatching(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing 3D discretized simulation, plane wave illumination...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

    def test_check_dimensions_verifications(self):
        for device in self.devices:
            # ignore warnings temporarily:
            warnings.simplefilter("ignore")
            
            wavelengths = torch.linspace(600.0, 700.0, 2)
            # - some dummy material
            material = tg.materials.MatConstant(3.0)

            # - fields
            e02d = tg.env.freespace_2d.PlaneWave()
            e03d = tg.env.freespace_3d.PlaneWave()

            # - environments
            env2d = tg.env.EnvHomogeneous2D(1.0)
            env3d = tg.env.EnvHomogeneous3D(1.0)

            # - structures
            # 2D
            s2d_mesh = tg.struct.StructDiscretizedSquare2D(
                discretization_config=tg.struct.surface_2d.square(2),
                step=10,
                materials=material,
            )
            try:
                s2d_line = s2d_mesh.convert_to_effective_polarizability_pair(
                    environment=env3d,
                    wavelengths=wavelengths,
                    verbose=False,
                    progress_bar=False,
                )
                raise Exception("ERROR! EXTRACT 2D with 3D ENV SHOULD RAISE AN ERROR!!")
            except ValueError:
                pass
            s2d_line = s2d_mesh.convert_to_effective_polarizability_pair(
                environment=env2d,
                wavelengths=wavelengths,
                verbose=False,
                progress_bar=False,
            )

            # 3D
            s3d_mesh = tg.struct.StructDiscretizedCubic3D(
                discretization_config=tg.struct.volume.cube(2),
                step=10,
                materials=tg.materials.MatConstant(2.25),
            )
            try:
                s3d_point = s3d_mesh.convert_to_effective_polarizability_pair(
                    environment=env2d,
                    wavelengths=wavelengths,
                    verbose=False,
                    progress_bar=False,
                )
                raise Exception("ERROR! EXTRACT 3D with 2D ENV SHOULD RAISE AN ERROR!!")
            except ValueError:
                pass
            s3d_point = s3d_mesh.convert_to_effective_polarizability_pair(
                environment=env3d,
                wavelengths=wavelengths,
                verbose=False,
                progress_bar=False,
            )
            
            test_cases = [
                [s2d_mesh, e02d, env2d, True],
                [s2d_line, e02d, env2d, True],
                [s2d_mesh, [e02d, e02d], env2d, True],
                [s3d_mesh, e03d, env3d, True],
                [s3d_point, e03d, env3d, True],
                [[s3d_mesh, s3d_point], [e03d, e03d], env3d, True],
                [s2d_line, e02d, env3d, False],
                [s3d_mesh, [e03d, e02d], env3d, False],
                [s3d_point, e03d, env2d, False],
                [[s2d_line, s2d_mesh], [e03d, e02d], env2d, False],
            ]

            # 3D Mie
            try:
                # ignore import warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    import treams
                mie_test = True
            except ModuleNotFoundError:
                print(
                    "Skipping test. Mie tests require package `treams`. "
                    + "Please install via `pip install treams`."
                )
                mie_test = False
            if mie_test:
                try:
                    s3d_mie = tg.struct.StructMieSphereEffPola3D(
                        wavelengths, environment=env2d, materials=material, radii=100.0
                    )
                    raise Exception(
                        "ERROR! EXTRACT MIE (3D) with 2D ENV SHOULD RAISE AN ERROR!!"
                    )
                except ValueError:
                    pass
                s3d_mie = tg.struct.StructMieSphereEffPola3D(
                    wavelengths, environment=env3d, materials=material, radii=100.0
                )
            if mie_test:
                test_cases += [
                [s3d_mie, e03d, env3d, True],
                [s3d_mie, e02d, env3d, False],
                ]


            for i, _t in enumerate(test_cases):
                try:
                    sim = tg.Simulation(
                        structures=_t[0],
                        illumination_fields=_t[1],
                        environment=_t[2],
                        wavelengths=wavelengths,
                        device=device,
                    )
                    if not _t[3]:
                        raise Exception(
                            "Simulation dimensionality check passed but "
                            + f"should have failed on case {i}."
                        )
                except ValueError:
                    if _t[3] == False:
                        pass  # everything ok: An exception is expected in this config
                    else:
                        raise Exception(
                            "Simulation dimensionality check failed but "
                            + f"should have passed on case {i}."
                        )
            if self.verbose:
                print("  - {}: dimension check test passed.".format(device))


# %%
if __name__ == "__main__":
    print("testing simulation dimensionality check...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)
